/* -*- tab-width: 4 -*- */

/* Input functions and primitives; the Scheme reader/parser.
 */

#include "global.h"
#include "elkkernel.h"

#include <ctype.h>
#include <limits.h>

#ifdef FLUSH_TIOCFLUSH
#  include <sys/ioctl.h>
#else
#  ifdef FLUSH_TCFLSH
#    include <termio.h>
#  endif
#endif

#ifdef FIONREAD_H
#  include FIONREAD_H
#endif

extern CHAR *index();
int GetStdIn(void);
int String_Getc (ELKObject port);
void String_Ungetc (ELKObject port, int c);
int Skip_Comment (ELKObject port);
void UnGetStdIn(long ch);
int Reader_Getc(int str, ELKObject port, FILE *f);
void Reader_Ungetc(int str, int c, ELKObject port, FILE *f);

ELKObject Sym_Quote,
	   Sym_Quasiquote,
	   Sym_Unquote,
	   Sym_Unquote_Splicing;

#define Octal(c) ((c) >= '0' && (c) <= '7')

static READFUN Readers[256];

static CHAR *Read_Buf;
static int Read_Size, Read_Max;

#define Read_Reset()   (Read_Size = 0)
#define Read_Store(c)  (Read_Size == Read_Max ? \
	(Read_Grow(), Read_Buf[Read_Size++] = (c)) : (Read_Buf[Read_Size++] = (c)))

static void Read_Grow (void) {
	Read_Max *= 2;
	Read_Buf = Safe_Realloc (Read_Buf, Read_Max);
}

ELKObject General_Read(ELKObject,int), Read_Sequence(ELKObject port, int vec, int konst);
ELKObject Read_Atom(ELKObject port, int konst), Read_Special(ELKObject port, int konst);
ELKObject Read_String(ELKObject port, int konst), Read_Sharp(ELKObject port, int konst);
ELKObject Read_True (ELKObject port, int chr, int konst);
ELKObject Read_False (ELKObject port, int chr, int konst);
ELKObject Read_Void (ELKObject port, int chr, int konst);
ELKObject Read_Hack (ELKObject port, int chr, int konst);
ELKObject Read_Vector (ELKObject port, int chr, int konst);
ELKObject Read_Radix (ELKObject port, int chr, int konst);
ELKObject Read_Char (ELKObject port, int chr, int konst);

void Init_Read (void) {
	Define_Symbol (&Sym_Quote, x_("quote"));
	Define_Symbol (&Sym_Quasiquote, x_("quasiquote"));
	Define_Symbol (&Sym_Unquote, x_("unquote"));
	Define_Symbol (&Sym_Unquote_Splicing, x_("unquote-splicing"));

	Readers['t'] = Readers['T'] = Read_True;
	Readers['f'] = Readers['F'] = Read_False;
	Readers['v'] = Readers['V'] = Read_Void;
	Readers['!'] = Read_Hack;  /* for interpreter files */
	Readers['('] = Read_Vector;
	Readers['b'] = Readers['B'] =
	Readers['o'] = Readers['O'] =
	Readers['d'] = Readers['D'] =
	Readers['x'] = Readers['X'] =
	Readers['e'] = Readers['E'] =
	Readers['i'] = Readers['I'] = Read_Radix;
	Readers['\\'] = Read_Char;

	Read_Max = 128;
	Read_Buf = Safe_Malloc (Read_Max);
}

/*********************** ADDED BY SMR FOR ELECTRIC INPUT ***********************/

CHAR OutBuf[256], *OutPtr = OutBuf, InBuf[256], *InPtr = InBuf, UnGetCh = 0;

int GetStdIn(void)
{
	int ch;
	CHAR *retstr;

	if (UnGetCh != 0)
	{
		ch = UnGetCh;
		UnGetCh = 0;
		return(ch);
	}

	if (InPtr == InBuf)
	{
		if (OutPtr != OutBuf)
		{
			/* pending output: use it as a prompt */
			*OutPtr = 0;
			retstr = ttygetlinemessages(OutBuf);
			OutPtr = OutBuf;
			if (retstr == 0) return(-1);
			estrcpy(InBuf, retstr);
		} else
		{
			/* just get a line */
			retstr = ttygetlinemessages(x_(""));
			if (retstr == 0) return(-1);
			estrcpy(InBuf, retstr);
		}
	}

	ch = *InPtr++;
	if (ch == 0)
	{
		ch = '\n';
		InPtr = InBuf;
	}
	return(ch);
}

void UnGetStdIn(long ch)
{
	UnGetCh = (CHAR)ch;
}

void DumpStdOut(CHAR *str, long len)
{
	int i;

	for(i=0; i<len; i++)
	{
		if (str[i] == '\n' || str[i] == '\r')
		{
			*OutPtr = 0;
			ttyputmsg(x_("%s"), OutBuf);
			OutPtr = OutBuf;
			continue;
		}
		*OutPtr++ = str[i];
	}
}

int Reader_Getc(int str, ELKObject port, FILE *f)
{
	int c;

	if (str)
	{
		c = String_Getc (port);
	} else
	{
		if (f == stdin)
		{
			c = GetStdIn();
		} else
		{
			c = getc(f);
		}
	}
	if (c == '\n') PORT(port)->lno++;
	return(c);
}

void Reader_Ungetc(int str, int c, ELKObject port, FILE *f)
{
	if (str)
	{
		String_Ungetc(port, c);
	} else
	{
		if (f == stdin)
		{
			UnGetStdIn(c);
		} else
		{
			(void)ungetc(c, f);
		}
	}
	if (c == '\n')
	{
		if (PORT(port)->lno > 1) PORT(port)->lno--;
	}
}

/*********************** END OF SMR ADDITIONS ***********************/

int String_Getc (ELKObject port) {
	register struct S_Port *p;
	register struct S_String *s;

	p = PORT(port);
	if (p->flags & P_UNREAD) {
		p->flags &= ~P_UNREAD;
		return p->unread;
	}
	s = STRING(p->name);
	return p->ptr >= s->size ? EOF : s->data[p->ptr++];
}

void String_Ungetc (ELKObject port, register int c) {
	PORT(port)->flags |= P_UNREAD;
	PORT(port)->unread = c;
}

void Check_Input_Port (ELKObject port) {
	Check_Type (port, T_Port);
	if (!(PORT(port)->flags & P_OPEN))
		Primitive_Error (x_("port has been closed: ~s"), port);
	if (!IS_INPUT(port))
		Primitive_Error (x_("not an input port: ~s"), port);
}

ELKObject P_Clear_Input_Port (int argc, ELKObject *argv) {
	Discard_Input (argc == 1 ? argv[0] : Curr_Input_Port);
	return Void;
}

void Discard_Input (ELKObject port) {
	register FILE *f;

	Check_Input_Port (port);
	if (PORT(port)->flags & P_STRING)
		return;
	f = PORT(port)->file;
#ifdef FLUSH_FPURGE
	(void)fpurge (f);
#else
#  ifdef FLUSH_BSD
	f->_cnt = 0;
	f->_ptr = f->_base;
#  endif
#endif
#ifdef FLUSH_TIOCFLUSH
	(void)ioctl (fileno (f), TIOCFLUSH, (CHAR *)0);
#else
#  ifdef FLUSH_TCFLSH
	(void)ioctl (fileno (f), TCFLSH, (CHAR *)0);
#  endif
#endif
}

ELKObject P_Unread_Char (int argc, ELKObject *argv)
{
	ELKObject port, ch;
	register struct S_Port *p;
	FILE *io;

	ch = argv[0];
	Check_Type (ch, T_Character);
	port = argc == 2 ? argv[1] : Curr_Input_Port;
	Check_Input_Port (port);
	p = PORT(port);
	if (p->flags & P_STRING)
	{
		if (p->flags & P_UNREAD)
			Primitive_Error (x_("cannot push back more than one char"));
		String_Ungetc (port, ELKCHAR(ch));
	} else
	{
		io = p->file;
		if (io == stdin)
		{
			UnGetStdIn(ELKCHAR(ch));
		} else
		{
			if (ungetc (ELKCHAR(ch), io) == EOF)
				Primitive_Error (x_("failed to push back char"));
		}
	}
	if (ELKCHAR(ch) == '\n' && PORT(port)->lno > 1) PORT(port)->lno--;
	return ch;
}

ELKObject P_Read_Char (int argc, ELKObject *argv) {
	ELKObject port;
	register FILE *f;
	register int c, str, flags;

	port = argc == 1 ? argv[0] : Curr_Input_Port;
	Check_Input_Port (port);
	f = PORT(port)->file;
	flags = PORT(port)->flags;
	str = flags & P_STRING;
	c = Reader_Getc(str, port, f);
	Reader_Tweak_Stream;
	return c == EOF ? Eof : Make_Char (c);
}

ELKObject P_Peek_Char (int argc, ELKObject *argv) {
	ELKObject a[2];

	a[0] = P_Read_Char (argc, argv);
	if (argc == 1)
		a[1] = argv[0];
	return EQ(a[0], Eof) ? Eof : P_Unread_Char (argc+1, a);
}

/* char-ready? cannot be implemented correctly based on FILE pointers.
 * The following is only an approximation; even if FIONREAD is supported,
 * the primitive may return #f although a call to read-char would not block.
 */
ELKObject P_Char_Readyp (int argc, ELKObject *argv) {
	ELKObject port;

	port = argc == 1 ? argv[0] : Curr_Input_Port;
	Check_Input_Port (port);
	if (PORT(port)->flags & P_STRING || feof (PORT(port)->file))
		return Trueob;
#ifndef WIN32
#  ifdef FIONREAD
	{
		long num = 0;
		(void)ioctl (fileno (PORT(port)->file), FIONREAD, (CHAR *)&num);
		if (num != 0)
			return Trueob;
	}
#  endif
#endif
	return Falseob;
}

ELKObject P_Read_String (int argc, ELKObject *argv) {
	ELKObject port;
	register FILE *f;
	register int c, str;

	port = argc == 1 ? argv[0] : Curr_Input_Port;
	Check_Input_Port (port);
	f = PORT(port)->file;
	str = PORT(port)->flags & P_STRING;
	Read_Reset ();
	while (1) {
		c = Reader_Getc(str, port, f);
		if (c == '\n' || c == EOF)
			break;
		Read_Store (c);
	}
	Reader_Tweak_Stream;
	return c == EOF ? Eof : Make_String (Read_Buf, Read_Size);
}

ELKObject P_Read (int argc, ELKObject *argv) {
	return General_Read (argc == 1 ? argv[0] : Curr_Input_Port, 0);
}

ELKObject General_Read (ELKObject port, int konst) {
	register FILE *f;
	register int c, str;
	ELKObject ret;

	Check_Input_Port (port);
	Flush_Output (Curr_Output_Port);
	f = PORT(port)->file;
	str = PORT(port)->flags & P_STRING;
	while (1) {
		c = Reader_Getc(str, port, f);
		if (c == EOF) {
			ret = Eof;
			break;
		}
		if (Whitespace (c))
			continue;
		if (c == ';') {
comment:
			if (Skip_Comment (port) == EOF) {
				ret = Eof;
				break;
			}
			continue;
		}
		if (c == '(') {
			ret = Read_Sequence (port, 0, konst);
		} else if (c == '#') {
			ret = Read_Sharp (port, konst);
			if (TYPE(ret) == T_Special)	  /* it was a #! */
				goto comment;
		} else {
			Reader_Ungetc(str, c, port, f);
			ret = Read_Atom (port, konst);
		}
		break;
	}
	Reader_Tweak_Stream;
	return ret;
}

int Skip_Comment (ELKObject port) {
	register FILE *f;
	register int c, str;

	f = PORT(port)->file;
	str = PORT(port)->flags & P_STRING;
	do {
		c = Reader_Getc(str, port, f);
	} while (c != '\n' && c != EOF);
	return c;
}

ELKObject Read_Atom (ELKObject port, int konst) {
	ELKObject ret;

	ret = Read_Special (port, konst);
	if (TYPE(ret) == T_Special)
		Reader_Error (port, x_("syntax error"));
	return ret;
}

ELKObject Read_Special (ELKObject port, int konst) {
	ELKObject ret;
	register int c, str;
	register FILE *f;

#define READ_QUOTE(sym) \
	( ret = Read_Atom (port, konst),\
	  konst ? (ret = Const_Cons (ret, Null), Const_Cons (sym, ret))\
		   : (ret = Cons (ret, Null), Cons (sym, ret)))

	f = PORT(port)->file;
	str = PORT(port)->flags & P_STRING;
again:
	c = Reader_Getc(str, port, f);
	switch (c) {
	case EOF:
eof:
		Reader_Tweak_Stream;
		Reader_Error (port, x_("premature end of file"));
	case ';':
		if (Skip_Comment (port) == EOF)
			goto eof;
		goto again;
	case ')':
		SET(ret, T_Special, c);
		return ret;
	case '(':
		return Read_Sequence (port, 0, konst);
	case '\'':
		return READ_QUOTE(Sym_Quote);
	case '`':
		return READ_QUOTE(Sym_Quasiquote);
	case ',':
		c = Reader_Getc(str, port, f);
		if (c == EOF)
			goto eof;
		if (c == '@') {
			return READ_QUOTE(Sym_Unquote_Splicing);
		} else {
			Reader_Ungetc(str, c, port, f);
			return READ_QUOTE(Sym_Unquote);
		}
	case '"':
		return Read_String (port, konst);
	case '#':
		ret = Read_Sharp (port, konst);
		if (TYPE(ret) == T_Special)
			goto again;
		return ret;
	default:
		if (Whitespace (c))
			goto again;
		Read_Reset ();
		if (c == '.') {
			c = Reader_Getc(str, port, f);
			if (c == EOF)
				goto eof;
			if (Whitespace (c)) {
				Reader_Ungetc(str, c, port, f);
				SET(ret, T_Special, '.');
				return ret;
			}
			Read_Store ('.');
		}
		while (!Whitespace (c) && !Delimiter (c) && c != EOF) {
			if (c == '\\') {
				c = Reader_Getc(str, port, f);
				if (c == EOF)
					break;
			}
			Read_Store (c);
			c = Reader_Getc(str, port, f);
		}
		Read_Store ('\0');
		if (c != EOF)
			Reader_Ungetc(str, c, port, f);
		ret = Parse_Number (port, Read_Buf, 10);
		if (Nullp (ret))
			ret = Intern (Read_Buf);
		return ret;
	}
	/*NOTREACHED*/
}

ELKObject Read_Sequence (ELKObject port, int vec, int konst) {
	ELKObject ret, e, tail, t;
	GC_Node3;

	ret = tail = Null;
	GC_Link3 (ret, tail, port);
	while (1) {
		e = Read_Special (port, konst);
		if (TYPE(e) == T_Special) {
			if (ELKCHAR(e) == ')') {
				GC_Unlink;
				return ret;
			}
			if (vec)
				Reader_Error (port, x_("wrong syntax in vector"));
			if (ELKCHAR(e) == '.') {
				if (Nullp (tail)) {
					ret = Read_Atom (port, konst);
				} else {
					e = Read_Atom (port, konst);
					/*
					 * Possibly modifying pure cons.  Must be fixed!
					 */
					Cdr (tail) = e;
				}
				e = Read_Special (port, konst);
				if (TYPE(e) == T_Special && ELKCHAR(e) == ')') {
					GC_Unlink;
					return ret;
				}
				Reader_Error (port, x_("dot in wrong context"));
			}
			Reader_Error (port, x_("syntax error"));
		}
		if (konst) t = Const_Cons (e, Null); else t = Cons (e, Null);
		if (!Nullp (tail))
			/*
			 * Possibly modifying pure cons.  Must be fixed!
			 */
			Cdr (tail) = t;
		else
			ret = t;
		tail = t;
	}
	/*NOTREACHED*/
}

ELKObject Read_String (ELKObject port, int konst) {
	register FILE *f;
	register int n, c, oc, str;

	Read_Reset ();
	f = PORT(port)->file;
	str = PORT(port)->flags & P_STRING;
	while (1) {
		c = Reader_Getc(str, port, f);
		if (c == EOF) {
eof:
			Reader_Tweak_Stream;
			Reader_Error (port, x_("end of file in string"));
		}
		if (c == '\\') {
			c = Reader_Getc(str, port, f);
			switch (c) {
			case EOF: goto eof;
			case 'b': c = '\b'; break;
			case 't': c = '\t'; break;
			case 'r': c = '\r'; break;
			case 'n': c = '\n'; break;
			case '0': case '1': case '2': case '3':
			case '4': case '5': case '6': case '7':
				oc = n = 0;
				do {
					oc <<= 3; oc += c - '0';
					c = Reader_Getc(str, port, f);
					if (c == EOF) goto eof;
				} while (Octal (c) && ++n <= 2);
				Reader_Ungetc(str, c, port, f);
				c = oc;
			}
		} else if (c == '"')
			break;
		Read_Store (c);
	}
	return General_Make_String (Read_Buf, Read_Size, konst);
}

ELKObject Read_Sharp (ELKObject port, int konst) {
	int c, str;
	FILE *f;
	CHAR buf[32];

	f = PORT(port)->file;
	str = PORT(port)->flags & P_STRING;
	c = Reader_Getc(str, port, f);
	if (c == EOF)
		Reader_Sharp_Eof;
	if (!Readers[c]) {
		esnprintf (buf, 32, x_("no reader for syntax #%c"), c);
		Reader_Error (port, buf);
	}
	return Readers[c](port, c, konst);
}

/*ARGSUSED*/
ELKObject Read_True (ELKObject port, int chr, int konst) {
	return Trueob;
}

/*ARGSUSED*/
ELKObject Read_False (ELKObject port, int chr, int konst) {
	return Falseob;
}

/*ARGSUSED*/
ELKObject Read_Void (ELKObject port, int chr, int konst) {
	ELKObject ret;

	ret = Const_Cons (Void, Null);
	return Const_Cons (Sym_Quote, ret);
}

/*ARGSUSED*/
ELKObject Read_Hack (ELKObject port, int chr, int konst) {
	return Special;
}

/*ARGSUSED*/
ELKObject Read_Vector (ELKObject port, int chr, int konst) {
	return List_To_Vector (Read_Sequence (port, 1, konst), konst);
}

/*ARGSUSED*/
ELKObject Read_Radix (ELKObject port, int chr, int konst) {
	int c, str;
	FILE *f;
	ELKObject ret;

	Read_Reset ();
	f = PORT(port)->file;
	str = PORT(port)->flags & P_STRING;
	Read_Store ('#'); Read_Store (chr);
	while (1) {
		c = Reader_Getc(str, port, f);
		if (c == EOF)
			Reader_Sharp_Eof;
		if (Whitespace (c) || Delimiter (c))
			break;
		Read_Store (c);
	}
	Reader_Ungetc(str, c, port, f);
	Read_Store ('\0');
	ret = Parse_Number (port, Read_Buf, 10);
	if (Nullp (ret))
		Reader_Error (port, x_("radix not followed by a valid number"));
	return ret;
}

/*ARGSUSED*/
ELKObject Read_Char (ELKObject port, int chr, int konst) {
	int c, str;
	FILE *f;
	CHAR buf[10], *p = buf;

	f = PORT(port)->file;
	str = PORT(port)->flags & P_STRING;
	c = Reader_Getc(str, port, f);
	if (c == EOF)
		Reader_Sharp_Eof;
	*p++ = c;
	while (1) {
		c = Reader_Getc(str, port, f);
		if (c == EOF)
			Reader_Sharp_Eof;
		if (Whitespace (c) || Delimiter (c))
			break;
		if (p == buf+9)
			Reader_Error (port, x_("syntax error in character constant"));
		*p++ = c;
	}
	Reader_Ungetc(str, c, port, f);
	if (p == buf+1)
		return Make_Char (*buf);
	*p = '\0';
	if (p == buf+3) {
		for (c = 0, p = buf; p < buf+3 && Octal (*p); p++)
			c = c << 3 | (*p - '0');
		if (p == buf+3)
			return Make_Char (c);
	}
	for (p = buf; *p; p++)
		if (isupper (*p))
			*p = tolower (*p);
	if (estrcmp (buf, x_("space")) == 0)
		return Make_Char (' ');
	if (estrcmp (buf, x_("newline")) == 0)
		return Make_Char ('\n');
	if (estrcmp (buf, x_("return")) == 0)
		return Make_Char ('\r');
	if (estrcmp (buf, x_("tab")) == 0)
		return Make_Char ('\t');
	if (estrcmp (buf, x_("formfeed")) == 0)
		return Make_Char ('\f');
	if (estrcmp (buf, x_("backspace")) == 0)
		return Make_Char ('\b');
	Reader_Error (port, x_("syntax error in character constant"));
	/*NOTREACHED*/
	return(Null);
}

void Define_Reader (int c, READFUN fun) {
	if (Readers[c] && Readers[c] != fun)
		Primitive_Error (x_("reader for `~a' already defined"), Make_Char (c));
	Readers[c] = fun;
}

ELKObject Parse_Number (ELKObject port, const CHAR *buf, int radix) {
	const CHAR *p;
	int c, i;
	int mdigit = 0, edigit = 0, expo = 0, neg = 0, point = 0;
	int gotradix = 0, exact = 0, inexact = 0;
	unsigned max;
	int maxdig;
	ELKObject ret;

	for ( ; *buf == '#'; buf++) {
		switch (*++buf) {
		case 'b': case 'B':
			if (gotradix++) return Null;
			radix = 2;
			break;
		case 'o': case 'O':
			if (gotradix++) return Null;
			radix = 8;
			break;
		case 'd': case 'D':
			if (gotradix++) return Null;
			radix = 10;
			break;
		case 'x': case 'X':
			if (gotradix++) return Null;
			radix = 16;
			break;
		case 'e': case 'E':
			if (exact++ || inexact) return Null;
			break;
		case 'i': case 'I':
			if (inexact++ || exact) return Null;
			break;
		default:
			return Null;
		}
	}
	p = buf;
	if (*p == '+' || (neg = *p == '-'))
		p++;
	for ( ; (c = *p) != 0; p++) {
		if (c == '.') {
			if (expo || point++)
				return Null;
		} else if (radix != 16 && (c == 'e' || c == 'E')) {
			if (expo++)
				return Null;
			if (p[1] == '+' || p[1] == '-')
				p++;
		} else if (radix == 16 && !index (x_("0123456789abcdefABCDEF"), c)) {
			return Null;
		} else if (radix < 16 && (c < '0' || c > '0' + radix-1)) {
			return Null;
		} else {
			if (expo) edigit++; else mdigit++;
		}
	}
	if (!mdigit || (expo && !edigit))
		return Null;
	if (point || expo) {
		if (radix != 10) {
			if (Nullp (port))
				return Null;
			Reader_Error (port, x_("reals must be given in decimal"));
		}
		/* Lacking ratnums, there's nothing we can do if #e has been
		 * specified-- just return the inexact number.
		 */
		return Make_Flonum (eatof (buf));
	}
#if 0		/* compiler hack */
	/*
	 * INT_MIN is a value with 1 in the sign bit and zeros elsewhere.
	 * Negating it implies reversing the bits and adding 1, which gives
	 * the same value.  However, some compilers notice that it is an overflow.
	 * Therefore, complementing the bits gives the proper intent: the largest
	 * integer.
	 *     -smr
	 */
	max = (neg ? (unsigned)(-INT_MIN) : INT_MAX);
#else
	max = (neg ? (unsigned)(~INT_MIN) : INT_MAX);
#endif
	maxdig = max % radix;
	max /= radix;
	for (i = 0, p = buf; (c = *p) != 0; p++) {
		if (c == '-' || c == '+') {
			buf++;
			continue;
		}
		if (radix == 16) {
			if (isupper (c))
				c = tolower (c);
			if (c >= 'a')
				c = '9' + c - 'a' + 1;
		}
		c -= '0';
		if (((unsigned)i > max || (unsigned)i == max) && c > maxdig) {
			ret = Make_Bignum (buf, neg, radix);
			return inexact ? Make_Flonum (Bignum_To_Double (ret)) : ret;
		}
		i *= radix; i += c;
	}
	if (neg)
		i = -i;
	return inexact ? Make_Flonum ((double)i) : Make_Integer (i);
}

void Reader_Error (ELKObject port, CHAR *msg) {
	CHAR buf[100];

	if (PORT(port)->flags & P_STRING) {
		esnprintf (buf, 100, x_("[string-port]: %u: %s"), PORT(port)->lno, msg);
		Primitive_Error (buf);
	} else {
		esnprintf (buf, 100, x_("~s: %u: %s"), PORT(port)->lno, msg);
		Primitive_Error (buf, PORT(port)->name);
	}
}
