/* -*- tab-width: 4 -*-
 *
 *     ********************************************************************* 
 *     * Copyright (C) 1988, 1990 Stanford University.                     * 
 *     * Permission to use, copy, modify, and distribute this              * 
 *     * software and its documentation for any purpose and without        * 
 *     * fee is hereby granted, provided that the above copyright          * 
 *     * notice appear in all copies.  Stanford University                 * 
 *     * makes no representations about the suitability of this            * 
 *     * software for any purpose.  It is provided "as is" without         * 
 *     * express or implied warranty.                                      * 
 *     ********************************************************************* 
 */

/*
 *	Routines to write and read history dump files
 */

#include "global.h"
#include "irsimnet.h"
#include "irsimglobals.h"


static	CHAR	fh_header[] = {x_("#HDUMP#\n")};	/* first line of dump file */
static	int	dumph_version = 4;


	/* define the number of bytes used to read/write history */
#define	NB_HEADER	(sizeof(fh_header) - 1)
#define	NB_NUMBER	4
#define	NB_NDINDEX	4
#define	NB_TIME		4
#ifdef HUGE_TIMETYPE
#define	NB_HTIME	8
#else
#define	NB_HTIME	4
#endif
#define	NB_PVAL		1
#define	NB_RTIME	2
#define	NB_DELAY	2
#define	NB_EVAL		1
#define	NB_VERSION	2

#define	MAGIC_NUM(DELTA, NNODES) (((DELTA) ^ (NNODES)) & 0xffff)

typedef struct
{
	CHAR  header[NB_HEADER];
	CHAR  hsize[NB_NUMBER];
	CHAR  nnodes[NB_NUMBER];
	CHAR  cur_delta[NB_HTIME];
	CHAR  magic[NB_NUMBER];
	CHAR  version[NB_VERSION];
	CHAR  time0[NB_HTIME];
} File_Head;

#define	Size_File_Head	(NB_HEADER + NB_NUMBER + NB_NUMBER + NB_HTIME + \
			NB_NUMBER + NB_VERSION + NB_HTIME)


typedef struct			/* format of history header, one per node */
{
	CHAR  node[NB_NDINDEX];		/* node for which history follows */
	CHAR  time[NB_TIME];		/* initial time (usually 0) */
	CHAR  pval[NB_PVAL];		/* initial packed-value */
} Node_Head;

#define	Size_Node_Head		(NB_NDINDEX + NB_TIME + NB_PVAL)


typedef struct		/* format of each history entry */
{
	CHAR  time[NB_TIME];		/* time of this change */
	CHAR  rtime[NB_RTIME];		/* input rise time */
	CHAR  delay[NB_DELAY];		/* associated delay */
	CHAR  pval[NB_PVAL];		/* packed-value (inp, val, punt) */
	CHAR  ptime[NB_DELAY];		/* punt time (only punted events) */
} File_Hist;

#define	Size_File_Hist		(NB_TIME + NB_RTIME + NB_DELAY + NB_PVAL)
#define	Size_Ptime		(NB_DELAY)
#define	Size_PuntFile_Hist	(Size_File_Hist + Size_Ptime)


	/* marks end of history.  Must be same size as Size_File_Hist! */
typedef struct
{
	CHAR  mark[NB_TIME];		/* marker, same for all nodes */
	CHAR  npend[NB_RTIME];		/* # of pending events */
	CHAR  dummy1[NB_DELAY];		/* just fill the space */
	CHAR  dummy2[NB_PVAL];
} EndHist;


typedef struct		/* format for pending events */
{
	CHAR  cause[NB_NDINDEX];
	CHAR  time[NB_TIME];
	CHAR  delay[NB_DELAY];
	CHAR  rtime[NB_RTIME];
	CHAR  eval[NB_EVAL];
} File_Pend;

#define	Size_File_Pend	(NB_NDINDEX + NB_TIME + NB_DELAY + NB_RTIME + NB_EVAL)


	/* macros to pack/unpack a value/inp/punt triplet from/to a byte */

#define	PACK_VAL(H)		((H->inp << 5) | (H->punt << 4) | H->val)
#define	GET_INP(PV)		((PV) >> 5 & 1)
#define	GET_VAL(PV)		((PV) & 0x7)
#define	IS_PUNT(PV)		((PV) & 0x10)


static	EndHist	h_end =
{
	{ '\0', '\0', '\0', '\040' },	/* mark */
	{ '\0', '\0' },					/* npend */
	{ '\0', '\0' },					/* dummy1 */
	{ '\n' }						/* dummy2 */
};


static int DumpNodeHist(nptr nd, Ulong ndindx, FILE *fp)
{
	register hptr  h;
	Node_Head      header;
	File_Hist      hist;

	if (nd->nflags & (POWER_RAIL | ALIAS | MERGED))
		return(0);

	irsim_PackBytes(header.node, ndindx, NB_NDINDEX);

	h = &(nd->head);
	irsim_PackTime(header.time, (TimeType)h->htime);
	irsim_PackBytes(header.pval, PACK_VAL(h), NB_PVAL);
	if (xfwrite((UCHAR1 *)&header, 1, Size_Node_Head, fp) <= 0)
		goto abort;

ttyputmsg(x_(""));
#ifdef HUGE_TIMETYPE
ttyputmsg(x_("%ld nodes, delta=%ld time=%lld htime=%lld"), irsim_nnodes, irsim_cur_delta, irsim_sim_time0, (TimeType)h->htime);
#else
ttyputmsg(x_("%ld nodes, delta=%ld time=%ld htime=%ld"), irsim_nnodes, irsim_cur_delta, irsim_sim_time0, (TimeType)h->htime);
#endif

	for(h = h->next; h != irsim_last_hist; h = h->next)
	{
#ifdef HUGE_TIMETYPE
ttyputmsg(x_("   Time %lld, punt=%d Delay=%d RTime=%d"), (TimeType)h->htime, h->punt, h->t.p.delay, h->t.p.rtime);
#else
ttyputmsg(x_("   Time %ld, punt=%d Delay=%d RTime=%d"), (TimeType)h->htime, h->punt, h->t.p.delay, h->t.p.rtime);
#endif
		irsim_PackTime(hist.time, (TimeType)h->htime);
		irsim_PackBytes(hist.pval, PACK_VAL(h), NB_PVAL);
		if (h->punt)
		{
			irsim_PackBytes(hist.delay, h->t.p.delay, NB_DELAY);
			irsim_PackBytes(hist.rtime, h->t.p.rtime, NB_RTIME);
			irsim_PackBytes(hist.ptime, h->t.p.ptime, NB_DELAY);
			if (xfwrite((UCHAR1 *)&hist, 1, Size_PuntFile_Hist, fp) <= 0)
				goto abort;
		} else
		{
			irsim_PackBytes(hist.delay, h->t.r.delay, NB_DELAY);
			irsim_PackBytes(hist.rtime, h->t.r.rtime, NB_RTIME);
			if (xfwrite((UCHAR1 *)&hist, 1, Size_File_Hist, fp) <= 0)
				goto abort;
		}
	}

	if (nd->events != NULL)
	{
		register evptr  ev;
		register int    n;
		File_Pend       pending;

		for(n = 0, ev = nd->events; ev != NULL; ev = ev->nlink, n++);
			irsim_PackBytes(h_end.npend, n, NB_RTIME);
ttyputmsg(x_("**Events: %d"), n);
		if (xfwrite((UCHAR1 *)&h_end, 1, Size_File_Hist, fp) <= 0)
			goto abort;
		for(ev = nd->events; ev != NULL; ev = ev->nlink)
		{
			ndindx = irsim_Node2index(ev->p.cause);
ttyputmsg(x_("  index %d: cause=%d time=%d delay=%d rtime=%d eval=%d"),
	ndindx, ev->ntime, ev->delay, ev->rtime, ev->eval);
			irsim_PackBytes(pending.cause, ndindx, NB_NDINDEX);
			irsim_PackTime(pending.time, ev->ntime);
			irsim_PackBytes(pending.delay, ev->delay, NB_DELAY);
			irsim_PackBytes(pending.rtime, ev->rtime, NB_RTIME);
			irsim_PackBytes(pending.eval, ev->eval, NB_EVAL);
			if (xfwrite((UCHAR1 *)&pending, 1, Size_File_Pend, fp) <= 0)
				goto abort;
		}
	} else
	{
ttyputmsg(x_("**Events: NONE"));
		irsim_PackBytes(h_end.npend, 0, NB_RTIME);
		if (xfwrite((UCHAR1 *)&h_end, 1, Size_File_Hist, fp) <= 0)
			goto abort;
	}

	return(0);

abort:
	ttyputmsg(x_("can't write to file, history dump aborted"));
	return(1);
}


static int WriteHistHeader(FILE *fd)
{
	File_Head  fh;
	long       mag;

	mag = MAGIC_NUM((long)irsim_cur_delta, irsim_nnodes);

	memcpy(fh.header, fh_header, NB_HEADER);
	irsim_PackBytes(fh.hsize, irsim_GetHashSize(), NB_NUMBER);
	irsim_PackBytes(fh.nnodes, irsim_nnodes, NB_NUMBER);
	irsim_PackTime(fh.cur_delta, irsim_cur_delta);
	irsim_PackBytes(fh.magic, mag, NB_NUMBER);
	irsim_PackBytes(fh.version, dumph_version, NB_VERSION);
	irsim_PackTime(fh.time0, irsim_sim_time0);
	if (xfwrite((UCHAR1 *)&fh, 1, Size_File_Head, fd) <= 0)
		return(-1);
	return(0);
}


static int ReadHistHead(FILE *fd, TimeType  *pNewTime, TimeType  *pTime0)
{
	File_Head  fh;
	int        n_nodes, hsize, n_version;
	TimeType   newTime, time0;
	long       magic;

	if (xfread((UCHAR1 *)&fh, 1, Size_File_Head, fd) != Size_File_Head)
	{
		ttyputmsg(x_("ReadHist: can't read file"));
		return(-1);
	}

	n_version = (int)irsim_UnpackBytes(fh.version, NB_VERSION);
	hsize = (int)irsim_UnpackBytes(fh.hsize, NB_NUMBER);
	n_nodes = (int)irsim_UnpackBytes(fh.nnodes, NB_NUMBER);
	newTime = irsim_UnpackTime(fh.cur_delta);
	magic = (long)irsim_UnpackBytes(fh.magic, NB_NUMBER);
	time0 = irsim_UnpackTime(fh.time0);

	if (estrncmp(fh_header, fh.header, NB_HEADER) != 0)
	{
		ttyputmsg(x_("ReadHist: not a history dump file"));
		return(-1);
	}

	if (n_version != dumph_version)
	{
		ttyputmsg(x_("ReadHist: Incompatible version: %d"), n_version);
		return(-1);
	}

	if ((irsim_nnodes != n_nodes && irsim_nnodes != 0) || hsize != irsim_GetHashSize() ||
	magic != MAGIC_NUM(newTime, n_nodes))
	{
		ttyputmsg(x_("ReadHist: incompatible or bad history dump"));
		return(-1);
	}

	*pNewTime = newTime;
	*pTime0 = time0;
	return(0);
}


void irsim_DumpHist(CHAR *fname)
{
	FILE  *fp;

	if ((fp = efopen(fname, x_("w"))) == NULL)
	{
		ttyputmsg(x_("can not open file '%s'"), fname);
		return;
	}

	if (WriteHistHeader(fp))
	{
		ttyputmsg(x_("can't write to file '%s'"), fname);
		(void)fclose(fp);
		return;
	}

	irsim_walk_net_index(DumpNodeHist, fp);

	(void)fclose(fp);
}


static int rd_hist(FILE  *fd, nptr  *pnlist)
{
	Node_Head     head;
	File_Hist     hist;
	EndHist       *pe;
	int           inp, val, n, pval, delay, rtime;
	Ulong         ndindx;
	nptr          nd, ndlist;
	TimeType      etime;
	struct Event  ev;

	pe = (EndHist *) &hist;
	ndlist = NULL;
	while(xfread((UCHAR1 *)&head, 1, Size_Node_Head, fd) == Size_Node_Head)
	{
		ndindx = (Ulong)irsim_UnpackBytes(head.node, NB_NDINDEX);
		if ((nd = irsim_Index2node(ndindx)) == NULL)
		{
			ttyputmsg(x_("history read aborted: could not find node"));
			*pnlist = ndlist;
			return(-1);
		}

		if (nd->nflags & (POWER_RAIL | ALIAS))
		{
			ttyputmsg(x_("warning: %s should not be in history"),
				pnode(nd));
		}

		etime = irsim_UnpackTime(head.time);
		pval = (int)irsim_UnpackBytes(head.pval, NB_PVAL);
		val = GET_VAL(pval);
		inp = GET_INP(pval);
		irsim_SetFirstHist(nd, val, inp, etime);

		nd->n.next = ndlist;
		ndlist = nd;

		if (nd->head.next != irsim_last_hist)
			irsim_FreeHistList(nd);

		for(;;)
		{
			if (xfread((UCHAR1 *)&hist, 1, Size_File_Hist, fd) != Size_File_Hist)
				goto badfile;
			if (memcmp(h_end.mark, pe->mark, NB_TIME) == 0)
				break;
			if (nd->nflags & (POWER_RAIL | ALIAS))
				continue;
			etime = irsim_UnpackTime(hist.time);
			delay = (int)irsim_UnpackBytes(hist.delay, NB_DELAY);
			rtime = (int)irsim_UnpackBytes(hist.rtime, NB_RTIME);
			pval = (int)irsim_UnpackBytes(hist.pval, NB_PVAL);
			val = GET_VAL(pval);
			inp = GET_INP(pval);

			if (IS_PUNT(pval))
			{
				if (xfread((UCHAR1 *)hist.ptime, 1, Size_Ptime, fd) != Size_Ptime)
					goto badfile;
				ev.eval = (UCHAR1)val;
				ev.ntime = etime;
				ev.delay = delay;
				ev.rtime = rtime;
				delay = (int)irsim_UnpackBytes(hist.ptime, NB_DELAY);
				etime -= (TimeType)delay;
				irsim_AddPunted(nd, &ev, etime);
			}
			else
				irsim_AddHist(nd, val, inp, etime, (long) delay, (long) rtime);
		}

		if (! (nd->nflags & POWER_RAIL))
		{
			nd->npot = nd->curr->val;
			if (nd->curr->inp)
			nd->nflags |= INPUT;
		}

		while(nd->events != NULL)	/* get rid of any pending events */
		irsim_free_event(nd->events);

		n = (int)irsim_UnpackBytes(pe->npend, NB_RTIME);
		while(n != 0)
		{
			File_Pend  pend;

			if (xfread((UCHAR1 *)&pend, 1, Size_File_Pend, fd) != Size_File_Pend)
				goto badfile;
			ndindx = (Ulong)irsim_UnpackBytes(pend.cause, NB_NDINDEX);
			etime = irsim_UnpackTime(pend.time);
			delay = (int)irsim_UnpackBytes(pend.delay, NB_DELAY);
			rtime = (int)irsim_UnpackBytes(pend.rtime, NB_RTIME);
			val = (int)irsim_UnpackBytes(pend.eval, NB_EVAL);

			irsim_cur_node = irsim_Index2node(ndindx);
			irsim_cur_delta = etime - (TimeType)delay;			/* fake the delay */
			irsim_enqueue_event(nd, val, (long)delay, (long) rtime);
			n--;
		}
	}

	*pnlist = ndlist;
	return(0);

badfile:
	ttyputmsg(x_("premature eof on history file"));
	*pnlist = ndlist;
	return(-1);
}


static void fix_transistors(register nptr nd)
{
	register lptr  l;

	while(nd != NULL)
	{
		for(l = nd->ngate; l != NULL; l = l->next)
			l->xtor->state = compute_trans_state(l->xtor);
		nd = nd->n.next;
	}
	for(l = irsim_VDD_node->ngate; l != NULL; l = l->next)
		l->xtor->state = compute_trans_state(l->xtor);
	for(l = irsim_GND_node->ngate; l != NULL; l = l->next)
		l->xtor->state = compute_trans_state(l->xtor);
}


void irsim_ReadHist(CHAR *fname)
{
	FILE       *fd;
	TimeType   newTime, time0;
	nptr       ndlist;

	if ((fd = efopen(fname, x_("r"))) == NULL)
	{
		ttyputmsg(x_("can not open file '%s'"), fname);
		return;
	}

	if (ReadHistHead(fd, &newTime, &time0))
	{
		(void)fclose(fd);
		return;
	}

	irsim_ClearInputs();

	if (rd_hist(fd, &ndlist))
	{
		nptr  n;

		for(n = ndlist; n != NULL; n = n->n.next)
		{
			irsim_FreeHistList(n);			/* undo any work done */
			while(n->events != NULL)
				irsim_free_event(n->events);
		}
	}
	else
	{
		irsim_sim_time0 = time0;
		irsim_cur_delta = newTime;
		if (irsim_cur_delta > 0)
			irsim_NoInit();
		if (irsim_VDD_node != NULL)
			fix_transistors(ndlist);
	}

	(void)fclose(fd);
}
