/* -*- tab-width: 4 -*-
 *
 *     ********************************************************************* 
 *     * Copyright (C) 1988, 1990 Stanford University.                     * 
 *     * Permission to use, copy, modify, and distribute this              * 
 *     * software and its documentation for any purpose and without        * 
 *     * fee is hereby granted, provided that the above copyright          * 
 *     * notice appear in all copies.  Stanford University                 * 
 *     * makes no representations about the suitability of this            * 
 *     * software for any purpose.  It is provided "as is" without         * 
 *     * express or implied warranty.                                      * 
 *     ********************************************************************* 
 */

#include "global.h"
#include "irsimnet.h"
#include "irsimglobals.h"

/* event driven mosfet simulator. Chris Terman (2/80) */

iptr  irsim_hinputs = NULL;	/* list of nodes to be driven high */
iptr  irsim_linputs = NULL;	/* list of nodes to be driven low */
iptr  irsim_uinputs = NULL;	/* list of nodes to be driven X */
iptr  irsim_xinputs = NULL;	/* list of nodes to be removed from input */

iptr  irsim_infree = NULL;	/* unused input list cells */

iptr  *irsim_listTbl[8];

void irsim_init_listTbl(void)
{
	int  i;

	for(i = 0; i < 8; i++) irsim_listTbl[i] = NULL;
	irsim_listTbl[INPUT_NUM(H_INPUT)] = &irsim_hinputs;
	irsim_listTbl[INPUT_NUM(L_INPUT)] = &irsim_linputs;
	irsim_listTbl[INPUT_NUM(U_INPUT)] = &irsim_uinputs;
	irsim_listTbl[INPUT_NUM(X_INPUT)] = &irsim_xinputs;

	/* initialize global variables */
	irsim_hinputs = NULL;
	irsim_linputs = NULL;
	irsim_uinputs = NULL;
	irsim_xinputs = NULL;
	irsim_infree = NULL;
}


#define	pvalue(node_name, node)	\
    formatinfstr(infstr, x_("%s=%c "), (node_name), x_("0XX1")[(node)->npot])


static void pgvalue(void *infstr, tptr t)
{
	register nptr  n;
	static CHAR    *states[] = { x_("OFF"), x_("ON"), x_("UKNOWN"), x_("WEAK") };

	if (irsim_debug)
		formatinfstr(infstr, x_("[%s] "), states[t->state]);

	if (t->ttype & GATELIST)
	{
		formatinfstr(infstr, x_("("));
		for(t = (tptr) t->gate; t != NULL; t = t->scache.t)
		{
			n = t->gate;
			pvalue(pnode(n), n);
		}

		formatinfstr(infstr, x_(") "));
	} else
	{
		n = t->gate;
		pvalue(pnode(n), n);
	}
}


static void pr_one_res(CHAR *s, double r)
{
	if (r < 1e-9 || r > 100e9)
		(void)esnprintf(s, 15, x_("%2.1e"), r);
	else
	{
		int  e = 3;

		if (r >= 1000.0)
			do { e++; r *= 0.001; } while(r >= 1000.0);
		else if (r < 1 && r > 0)
			do { e--; r *= 1000; } while(r < 1.0);
		(void)esnprintf(s, 15, x_("%.1f%c"), r, x_("num\0KMG")[e]);
	}
}


static void pr_t_res(void *infstr, Resists *r)
{
	CHAR    buf[3][15];

	pr_one_res(buf[0], r->rstatic);
	pr_one_res(buf[1], r->dynhigh);
	pr_one_res(buf[2], r->dynlow);
	formatinfstr(infstr, x_("[%s, %s, %s]"), buf[0], buf[1], buf[2]);
}


static void ptrans(void *infstr, tptr t)
{
	formatinfstr(infstr, x_("%s "), irsim_ttype[BASETYPE(t->ttype)]);
	if (BASETYPE(t->ttype) != RESIST)
		pgvalue(infstr, t);

	pvalue(pnode(t->source), t->source);
	pvalue(pnode(t->drain), t->drain);
	pr_t_res(infstr, t->r);
	if (t->tlink != t && (irsim_treport & REPORT_TCOORD))
		formatinfstr(infstr, x_(" <%d,%d>"), t->x.pos, t->y.pos);
}


void irsim_idelete(register nptr n, register iptr *list)
{
	register iptr  p = *list;
	register iptr  q;

	if (p == NULL)
		return;
	else if (p->inode == n)
	{
		*list = p->next;
		FreeInput(p);
	}
	else
	{
		for(q = p->next; q != NULL; p = q, q = p->next)
			if (q->inode == n)
		{
			p->next = q->next;
			FreeInput(q);
			return;
		}
	}
}


void irsim_iinsert(nptr n, iptr *list)
{
	register iptr  p = irsim_infree;

	if (p == NULL)
		p = irsim_infree = (iptr) irsim_MallocList(sizeof(struct Input), 1);
	irsim_infree = p->next;

	p->next = *list;
	*list = p;
	p->inode = n;
}


void irsim_iinsert_once(nptr n, iptr *list)
{
	register iptr  p;

	for(p = *list; p != NULL; p = p->next)
		if (p->inode == n)
			return;

	irsim_iinsert(n, list);
}


static	int clear_input(register nptr n)
{
	if (!(n->nflags & POWER_RAIL))
		n->nflags &= ~INPUT;
	return(0);
}


void irsim_ClearInputs(void)
{
	register int   i;
	register iptr  p, next;
	register nptr  n;

	for(i = 0; i < 5; i++)
	{
		if (irsim_listTbl[i] == NULL)
			continue;
		for(p = *irsim_listTbl[i]; p != NULL; p = next)
		{
			next = p->next;
			n = p->inode;
			FreeInput(p);
			if (!(n->nflags & POWER_RAIL))
				n->nflags &= ~(INPUT_MASK | INPUT);
		}
		*(irsim_listTbl[i]) = NULL;
	}
	irsim_walk_net(clear_input);
}


/*
 * set/clear input status of node and add/remove it to/from corresponding list.
 */
int irsim_setin(register nptr n, CHAR *which)
{
	while(n->nflags & ALIAS)
		n = n->nlink;

	if (n->nflags & (POWER_RAIL | MERGED))	/* Gnd, Vdd, or merged node */
	{
		if ((n->nflags & MERGED) || x_("lxuh")[n->npot] != *which)
			ttyputmsg(x_("Can't drive `%s' to `%c'"),
		pnode(n), *which);
	} else
	{
		iptr  *list = irsim_listTbl[INPUT_NUM(n->nflags)];

#	define	WASINP(N, P)	(((N)->nflags & INPUT) && (N)->npot == (P))

		switch(*which)
		{
			case 'h':
				if (list != NULL && list != &irsim_hinputs)
				{
					n->nflags = n->nflags & ~INPUT_MASK;
					irsim_idelete(n, list);
				}
				if (! (list == &irsim_hinputs || WASINP(n, HIGH)))
				{
					n->nflags = (n->nflags & ~INPUT_MASK) | H_INPUT;
					irsim_iinsert(n, &irsim_hinputs);
				}
				break;

			case 'l':
				if (list != NULL && list != &irsim_linputs)
				{
					n->nflags = n->nflags & ~INPUT_MASK;
					irsim_idelete(n, list);
				}
				if (! (list == &irsim_linputs || WASINP(n, LOW)))
				{
					n->nflags = (n->nflags & ~INPUT_MASK) | L_INPUT;
					irsim_iinsert(n, &irsim_linputs);
				}
				break;

			case 'u':
				if (list != NULL && list != &irsim_uinputs)
				{
					n->nflags = n->nflags & ~INPUT_MASK;
					irsim_idelete(n, list);
				}
				if (! (list == &irsim_uinputs || WASINP(n, X)))
				{
					n->nflags = (n->nflags & ~INPUT_MASK) | U_INPUT;
					irsim_iinsert(n, &irsim_uinputs);
				}
				break;

			case 'x':
				if (list == &irsim_xinputs)
					break;
				if (list)
				{
					n->nflags = n->nflags & ~INPUT_MASK;
					irsim_idelete(n, list);
				}
				if (n->nflags & INPUT)
				{
					n->nflags = (n->nflags & ~INPUT_MASK) | X_INPUT;
					irsim_iinsert(n, &irsim_xinputs);
				}
				break;

			default:
				return(0);
		}
	}
	return(1);
}


static int wr_value(nptr n, CHAR *fp)
{
	if (! (n->nflags & (ALIAS | POWER_RAIL)))
	{
		int ch = ((n->nflags & INPUT) ? '4' : '0') + n->npot;
		(void)putc(ch, (FILE *)fp);
	}
	return(0);
}


int irsim_wr_state(CHAR *fname)
{
	FILE  *fp;

	fp = xcreate(fname, el_filetypetext, 0, 0);
	if (fp == NULL) return(1);

	(void)efprintf(fp, x_("%d\n"), irsim_nnodes);
	irsim_walk_net1(wr_value, (CHAR *)fp);
	(void)fclose(fp);
	return(0);
}


typedef struct
{
	FILE  *file;
	int   errs;
	int   restore;
} StateFile;


static int rd_stvalue(nptr n, CHAR *starg)
{
	int   ch, inp;
	StateFile *st = (StateFile *)starg;

	if (n->nflags & (ALIAS | POWER_RAIL))
		return(0);

	irsim_FreeHistList(n);
	while(n->events != NULL)		/* remove any pending events */
		irsim_free_event(n->events);

	if (st->file == NULL)
	{
		ch = X;
		st->errs ++;
	} else
	{
		ch = getc(st->file);
		if (ch == EOF)
		{
			ch = X;
			st->errs ++;
			(void)fclose(st->file);
			st->file = NULL;
		} else if (ch < '0' || ch > '7' || ch == '2' || ch == '6')
		{
			st->errs ++;
			ch = X;
		} else if (st->restore && ch >= '4')
		{
			ch = ch - '4';
			inp = 1;
		} else
		{
			ch = (ch - '0') & (LOW | X | HIGH);
			inp = 0;
		}
	}

	if (n->nflags & MERGED)
		return(0);

	if (inp)
		n->nflags |= INPUT;

	n->head.val = ch;
	n->head.inp = inp;

	if (ch != n->npot)
	{
		register lptr  l;
		register tptr  t;

		n->npot = ch;
		for(l = n->ngate; l != NULL; l = l->next)
		{
			t = l->xtor;
			t->state = compute_trans_state(t);
		}
	}
	return(0);
}


CHAR *irsim_rd_state(CHAR *fname, int restore)
{
	CHAR       rline[25], *dummy;
	static CHAR retstring[100];
	StateFile  st;

	st.file = xopen(fname, el_filetypetext, 0, &dummy);
	if (st.file == NULL)
		return(x_("can not read state file"));

	(void)xfgets(rline, 25, st.file);
	if (eatoi(rline) != irsim_nnodes)
	{
		(void)fclose(st.file);
		return(x_("bad node count in state file"));
	}

	irsim_ClearInputs();

	if (irsim_analyzerON)
		irsim_StopAnalyzer();

	irsim_sim_time0 = irsim_cur_delta = 0;

	st.errs = 0;
	st.restore = restore;
	irsim_walk_net1(rd_stvalue, (CHAR *)&st);

	irsim_NoInit();

	if (irsim_analyzerON)
		irsim_RestartAnalyzer(irsim_sim_time0, irsim_cur_delta, FALSE);

	if (st.file == NULL)
	{
		(void)esnprintf(retstring, 100, x_("premature EOF in state file (%d errors)"),
			st.errs);
		return(retstring);
	}

	(void)fclose(st.file);

	if (st.errs != 0)
	{
		(void)esnprintf(retstring, 100, x_("%d errors found in state file"), st.errs);
		return(retstring);
	}
	return(NULL);
}

int irsim_info(register nptr n, CHAR *which)
{
	register tptr  t;
	register lptr  l;
	CHAR           *name;
	REGISTER void *infstr;

	if (n == NULL)
		return(0);

	if (stopping(STOPREASONSIMULATE))
		return(1);

	name = pnode(n);
	while(n->nflags & ALIAS)
		n = n->nlink;

	if (n->nflags & MERGED)
	{
		ttyputmsg(x_("%s => node is inside a transistor stack"), name);
		return(1);
	}

	infstr = initinfstr();
	pvalue(name, n);
	if (n->nflags & INPUT)
		formatinfstr(infstr, x_("[NOTE: node is an input] "));
	formatinfstr(infstr, x_("(vl=%.2f vh=%.2f) "), n->vlow, n->vhigh);
	if (n->nflags & USERDELAY)
		formatinfstr(infstr, x_("(tplh=%d, tphl=%d) "), n->tplh, n->tphl);
	formatinfstr(infstr, x_("(%5.4f pf) "), n->ncap);
	ttyputmsg(x_("%s"), returninfstr(infstr));

	infstr = initinfstr();
	if (*which == '?')
	{
		addstringtoinfstr(infstr, x_("is computed from:"));
		for(l = n->nterm; l != NULL && !stopping(STOPREASONSIMULATE); l = l->next)
		{
			t = l->xtor;
			addstringtoinfstr(infstr, x_("  "));
			if (irsim_debug == 0)
			{
				CHAR  *drive = NULL;
				nptr  rail;

				rail = (t->drain->nflags & POWER_RAIL) ? t->drain : t->source;
				if (BASETYPE(t->ttype) == NCHAN && rail == irsim_GND_node)
					drive = x_("pulled down by ");
				else if (BASETYPE(t->ttype) == PCHAN && rail == irsim_VDD_node)
					drive = x_("pulled up by ");
				else if (BASETYPE(t->ttype) == DEP && rail == irsim_VDD_node &&
					other_node(t, rail) == t->gate)
						drive = x_("pullup ");
				else
					ptrans(infstr, t);

				if (drive != NULL)
				{
					addstringtoinfstr(infstr, drive);
					pgvalue(infstr, t);
					pr_t_res(infstr, t->r);
					if (t->tlink != t && (irsim_treport & REPORT_TCOORD))
						formatinfstr(infstr, x_(" <%d,%d>"), t->x.pos, t->y.pos);
				}
			}
			else
				ptrans(infstr, t);
		}
	} else
	{
		addstringtoinfstr(infstr, x_("affects:"));
		for(l = n->ngate; l != NULL && !stopping(STOPREASONSIMULATE); l = l->next)
			ptrans(infstr, l->xtor);
	}
	ttyputmsg(x_("%s"), returninfstr(infstr));

	if (stopping(STOPREASONSIMULATE))
		ttyputmsg(x_("-- interrupted"));

	if (n->events != NULL)
	{
		register evptr  e;

		ttyputmsg(x_("Pending events:"));
		for(e = n->events; e != NULL; e = e->nlink)
			ttyputmsg(x_("   transition to %c at %2.2fns"),
				x_("0XX1")[e->eval], d2ns(e->ntime));
	}

	return(1);
}
