/* -*- tab-width: 4 -*-
 *
 *     ********************************************************************* 
 *     * Copyright (C) 1988, 1990 Stanford University.                     * 
 *     * Permission to use, copy, modify, and distribute this              * 
 *     * software and its documentation for any purpose and without        * 
 *     * fee is hereby granted, provided that the above copyright          * 
 *     * notice appear in all copies.  Stanford University                 * 
 *     * makes no representations about the suitability of this            * 
 *     * software for any purpose.  It is provided "as is" without         * 
 *     * express or implied warranty.                                      * 
 *     ********************************************************************* 
 */

#include "global.h"
#include "sim.h"
#include "irsimnet.h"
#include "irsimglobals.h"

/* useful subroutines for dealing with network */

CHAR   irsim_vchars[] = {x_("0XX1")};

CHAR  *irsim_ttype[NTTYPES] =
{
	x_("n-channel"),
	x_("p-channel"),
	x_("depletion"),
	x_("resistor")
#ifdef USER_SUBCKT
	,
	x_("reserved"),
	x_("subckt")
#endif
};


#define	HASHSIZE		4387	/* hash table size */
#define	NBIT_HASH		14	/* number of bits for HASHSIZE */

static	nptr  hash[HASHSIZE];


int irsim_GetHashSize(void)
{
	return(HASHSIZE);
}


/* hashing function used in interning symbols */
static	CHAR  lcase[128] = 
{
    0,		01,	02,	03,	04,	05,	06,	07,
    010,	011,	012,	013,	014,	015,	016,	017,
    020,	021,	022,	023,	024,	025,	026,	027,
    030,	031,	032,	033,	034,	035,	036,	037,
    ' ',	'!',	'"',	'#',	'$',	'%',	'&',	047,
    '(',	')',	'*',	'+',	',',	'-',	'.',	'/',
    '0',	'1',	'2',	'3',	'4',	'5',	'6',	'7',
    '8',	'9',	':',	';',	'<',	'=',	'>',	'?',
    '@',	'a',	'b',	'c',	'd',	'e',	'f',	'g',
    'h',	'i',	'j',	'k',	'l',	'm',	'n',	'o',
    'p',	'q',	'r',	's',	't',	'u',	'v',	'w',
    'x',	'y',	'z',	'[',	0134,	']',	'^',	'_',
    '`',	'a',	'b',	'c',	'd',	'e',	'f',	'g',
    'h',	'i',	'j',	'k',	'l',	'm',	'n',	'o',
    'p',	'q',	'r',	's',	't',	'u',	'v',	'w',
    'x',	'y',	'z',	'{',	'|',	'}',	'~',	0177
};



/*
 * Routine to free all memory allocated in this module.
 */
void irsim_freensubrsmemory(void)
{
	int i;
	nptr n;

	for(i=0; i<HASHSIZE; i++)
	{
		while (hash[i] != 0)
		{
			n = hash[i];
			hash[i] = n->hnext;
			efree((CHAR *)n->nname);
			efree((CHAR *)n);
		}
	}
	while (irsim_freeNodes != 0)
	{
		n = irsim_freeNodes;
		irsim_freeNodes = n->nlink;
		efree((CHAR *)n);
	}
}


static int sym_hash(register CHAR  *name)
{
	register int  hashcode = 0;

	do
		hashcode = (hashcode << 1) ^ (*name | 0x20);
	while(*(++name));
	return(((hashcode >= 0) ? hashcode : ~hashcode) % HASHSIZE);
}


/* compare pattern with string, case doesn't matter.  "*" wildcard accepted */
int irsim_str_match(register CHAR *p, register CHAR *s)
{
	for(;;)
	{
		if (*p == '*')
		{
			/* skip past multiple wildcards */
			do
				p++;
			while(*p == '*');

			/* if pattern ends with wild card, automatic match */
			if (*p == 0)
				return(1);

			/* *p now points to first non-wildcard character, find matching
			* character in string, then recursively match remaining pattern.
			* if recursive match fails, assume current '*' matches more...
			*/
			while(*s != 0)
			{
				while(lcase[*s] != lcase[*p])
					if (*s++ == 0)
						return(0);
				if (irsim_str_match(p + 1, ++s))
					return(1);
			}

			/* couldn't find matching character after '*', no match */
			return(0);
		}
		else if (*p == 0)
			return(*s == 0);
		else if (lcase[*p++] != lcase[*s++])
			break;
	}
	return(0);
}


/* find node in network */
nptr irsim_find(register CHAR *name)
{
	register nptr  ntemp;
	register int   cmp = 1;

	if (irsim_txt_coords && name[0] == '@' && name[1] == '=')
		if ((ntemp = irsim_FindNode_TxtorPos(name)) != NULL)
			return(ntemp);

	for(ntemp = hash[sym_hash(name)]; ntemp != NULL; ntemp = ntemp->hnext)
	{
		if ((cmp = namesame(name, ntemp->nname)) >= 0)
			break;
	}
	return((cmp == 0) ? ntemp : NULL);
}


/*
 * Get node structure.  If not found, create a new one.
 * If create is TRUE a new node is created and NOT entered into the table.
 */
static int warnVdd = FALSE, warnGnd = FALSE ;

nptr irsim_GetNode(register CHAR *name)
{
	register nptr  n, *prev;
	register int   i;
	int skip = 0;

	prev = &hash[sym_hash(name)];
	for(i = 1, n = *prev; n != NULL; n = *(prev = &n->hnext))
		if ((i = namesame(name, n->nname)) >= 0)
			break;

	if (i == 0)
	{
		if (estrcmp(name, n->nname) != 0)
		{
			if (namesame(name, x_("Vdd")) == 0)
			{
				skip = warnVdd ; warnVdd = TRUE ;
			}
			if (namesame(name, x_("GND")) == 0)
			{
				skip = warnGnd ; warnGnd = TRUE ;
			}
			if (skip == 0) 
				ttyputerr(x_("Warning: Aliasing nodes '%s' and '%s'"),
					name, n->nname);
		}
		while(n->nflags & ALIAS)
			n = n->nlink;
		return(n);
	}

	/* allocate new node from free storage */
	if ((n = (nptr) irsim_freeNodes) == NULL)
		n = (nptr) irsim_MallocList(sizeof(struct Node), 1);
	irsim_freeNodes = n->nlink;

	irsim_nnodes++;

	n->hnext = *prev;	/* insert node into hash table, after prev */
	*prev = n;

	/* initialize node entries */
	n->ngate = n->nterm = NULL;
#ifdef HIST_ONLY_TRACES
	n->nflags = HIST_OFF;
#else
	n->nflags = 0;
#endif
	n->ncap = (float)MIN_CAP;
	n->vlow = (float)irsim_LOWTHRESH;
	n->vhigh = (float)irsim_HIGHTHRESH;
	n->c.time = 0;
	n->tplh = 0;
	n->tphl = 0;
	n->t.cause = NULL;
	n->nlink = NULL;
	n->events = NULL;
	n->npot = X;
	n->awpending = NULL ;

	n->head.next = irsim_last_hist;
	n->head.htime = 0;
	n->head.val = X;
	n->head.inp = 0;
	n->head.punt = 0;
	n->head.t.r.rtime = n->head.t.r.delay = 0;
	n->curr = &(n->head);

	/* store all node-names as strings */
	i = estrlen(name) + 1;
	n->nname = (CHAR *)emalloc(i * SIZEOFCHAR, sim_tool->cluster);
	memcpy(n->nname, name, i);

	return(n);
}


nptr irsim_GetNewNode(CHAR *name)
{
	register nptr  n;
	int            i;

	if (irsim_VDD_node != NULL && namesame(name, pnode(irsim_VDD_node)) == 0)
		return(irsim_VDD_node);
	if (irsim_GND_node != NULL && namesame(name, pnode(irsim_GND_node)) == 0)
		return(irsim_GND_node);

	if ((n = (nptr) irsim_freeNodes) == NULL)
		n = (nptr) irsim_MallocList(sizeof(struct Node), 1);
	irsim_freeNodes = n->nlink;

	irsim_nnodes++;

	n->hnext = n;	/* node NOT inserted in hash-table */

	n->ngate = n->nterm = NULL;
	n->nflags = 0;
	n->ncap = (float)MIN_CAP;
	n->vlow = (float)irsim_LOWTHRESH;
	n->vhigh = (float)irsim_HIGHTHRESH;
	n->c.time = 0;
	n->tplh = 0;
	n->tphl = 0;
	n->t.cause = NULL;
	n->nlink = NULL;
	n->events = NULL;
	n->npot = X;

	n->head.next = irsim_last_hist;
	n->head.htime = 0;
	n->head.val = X;
	n->head.inp = 0;
	n->head.punt = 0;
	n->head.t.r.rtime = n->head.t.r.delay = 0;
	n->curr = &(n->head);

	/* store all node-names as strings */
	i = estrlen(name) + 1;
	n->nname = (CHAR *)emalloc(i * SIZEOFCHAR, sim_tool->cluster);
	memcpy(n->nname, name, i);

	return(n);
}


/* insert node into hash table.  Keep table entry sorted in ascending order */
void irsim_n_insert(register nptr nd)
{
	register nptr  n, *prev;
	register CHAR  *name;
	register int   i = 1;

	name = pnode(nd);
	prev = &hash[sym_hash(name)];
	for(n = *prev; n != NULL; n = *(prev = &n->hnext))
	{
		i = namesame(name, n->nname);
		if (i >= 0)
			break;
	}
	if (i == 0)
	{
		if (n != nd)
			ttyputmsg(x_("irsim_n_insert: multiple node '%s'"), pnode(nd));
		return;
	}
	nd->hnext = *prev;
	*prev = nd;
}


/*
 * Delete node from hash table, and free its name.
 */
void irsim_n_delete(register nptr nd)
{
	register nptr  n, *prev;

	prev = &hash[sym_hash(pnode(nd))];
	for(n = *prev; n != NULL ; n = *(prev = &n->hnext))
	{
		if (n == nd)
		{
			efree(n->nname);
			n->nname = NULL;
			*prev = n->hnext;
			n->hnext = n;		/* mark node not in hash-table */
			return;
		}
	}
}


/*
 * Visit each node in network, calling function passed.
 * If 'fun' returns anything <> 0 then terminate prematurely.
 */
void irsim_walk_net(int (*fun)(nptr))
{
	register int   index;
	register nptr  n;

	for(index = 0; index < HASHSIZE; index++)
		for(n = hash[index]; n; n = n->hnext)
			if ((*fun)(n))
				return;
}

/*
 * Visit each node in network, calling function passed.
 * If 'fun' returns anything <> 0 then terminate prematurely.
 */
void irsim_walk_net1(int (*fun)(nptr, CHAR*), CHAR *arg)
{
	register int   index;
	register nptr  n;

	for(index = 0; index < HASHSIZE; index++)
		for(n = hash[index]; n; n = n->hnext)
			if ((*fun)(n, arg))
				return;
}


/*
 * Visit each node in network, calling function passed with arguments:
 *	node	-> the current node.
 *	index	-> number that identifies node in the hash-table.
 *	arg	-> whatever argument the caller provided.
 * If 'fun' returns FALSE then terminate prematurely.
 */
void irsim_walk_net_index(int (*fun)(nptr, Ulong, FILE*), FILE *arg)
{
	register Uint  mi, ma;
	register nptr  n;

	for(ma = 0; ma < HASHSIZE; ma++)
		for(n = hash[ma], mi = 0; n; n = n->hnext, mi++)
	{
		Ulong  val = (mi << NBIT_HASH) | ma;
		if ((*fun)(n, val, arg))
			return;
	}
}


/*
 * Return a list of all nodes in the network.
 */
nptr irsim_GetNodeList(void)
{
	register int   index;
	register nptr  n, *last;
	struct Node    head;

	last = &(head.n.next);
	for(index = 0; index < HASHSIZE; index++)
	{
		for(n = hash[index]; n != NULL; n = n->hnext)
		{
			*last = n;
			last = &(n->n.next);
		}
	}
	*last = NULL;
	return(head.n.next);
}


/*
 * Return the node corresponding to the index number (returned by irsim_Node2index).
 */
nptr irsim_Index2node(Ulong index)
{
	register nptr  n;
	register unsigned  ma, mi;

	ma = index & ((1 << NBIT_HASH) - 1);
	mi = index >> NBIT_HASH;
	if (ma >= HASHSIZE)
		return(NULL);
	for(n = hash[ma]; n != NULL && mi != 0; n = n->hnext, mi--);
	return(n);
}


/*
 * Return a number that identifies the node in the name hash-table.
 */
Ulong irsim_Node2index(nptr nd)
{
	register nptr      n;
	register unsigned  mi, ma;
	Ulong              val;

	if (nd != NULL)
	{
		ma = sym_hash(pnode(nd));
		for(n = hash[ma], mi = 0; n != NULL; n = n->hnext, mi++)
		{
			if (n == nd)
				goto got_it;
		}
	}
	ma = HASHSIZE;
	mi = 0;

got_it:
	val = (mi << NBIT_HASH) | ma;
	return(val);
}


/* visit each node in network, calling function passed as arg with any node
 * whose name matches pattern
 */
int irsim_match_net(CHAR *pattern, int (*fun)(nptr, CHAR*), CHAR *arg)
{
	register int   index;
	register nptr  n;
	int            total = 0;

	for(index = 0; index < HASHSIZE; index++)
		for(n = hash[index]; n; n = n->hnext)
			if (irsim_str_match(pattern, pnode(n)))
				total += (*fun)(n, arg);

	return(total);
}


/* return pointer to asciz name of node */


/* initialize hash table */
void irsim_init_hash(void)
{
	register int  i;

	for(i = 0; i < HASHSIZE; i += 1)
		hash[i] = NULL;
}
