/* -*- tab-width: 4 -*-
 *
 *     ********************************************************************* 
 *     * Copyright (C) 1988, 1990 Stanford University.                     * 
 *     * Permission to use, copy, modify, and distribute this              * 
 *     * software and its documentation for any purpose and without        * 
 *     * fee is hereby granted, provided that the above copyright          * 
 *     * notice appear in all copies.  Stanford University                 * 
 *     * makes no representations about the suitability of this            * 
 *     * software for any purpose.  It is provided "as is" without         * 
 *     * express or implied warranty.                                      * 
 *     ********************************************************************* 
 */

#include "global.h"
#include "sim.h"
#include "irsimnet.h"
#include "irsimglobals.h"
#include "irsimnet_macros.h"

CHAR    irsim_version[] = {x_("version 9.5")};

/* front end for mos simulator -- Chris Terman (6/84) */
/* sunbstantial changes: Arturo Salz (88) */

#define	LSIZE       2000	/* max size of command line (in chars) */
#define	MAXARGS      100	/* maximum number of command-line arguments */
#define	CMDTBLSIZE    64	/* size of command hash-table */
#define	MAXCOL        80	/* maximum width of print line */

#define	ITERATOR_START	'{'
#define	ITERATOR_END	'}'

#define	SIZEOF(X)	((int) sizeof(X))


typedef struct _Cmd
{
	CHAR         *name;			/* name of this command */
	int          (*handler)();	/* handler for this command */
	short        nmin, nmax;	/* min and max number of arguments */
	CHAR         *help;			/* command description */
	struct _Cmd  *next;			/* list of commands in bucket */
} Command;


typedef struct _Path
{
	struct _Path  *next;
	CHAR          name[1];
} Path;


typedef struct sequence *sptr;

typedef struct sequence
{
	sptr    next;			/* next vector in linked list */
	int     which;			/* 0 => node; 1 => vector */
	union
	{
		nptr  n;
		bptr  b;
	} ptr;					/* pointer to node/vector */
	int     vsize;			/* size of each value */
	int     nvalues;		/* number of values specified */
	CHAR    values[1];		/* array of values */
} sequence;


static	Command	*cmdtbl[CMDTBLSIZE];	/* command hash-table */

static	Bits	*blist = NULL;		/* list of vectors */

static	sptr	slist = NULL;		/* list of sequences */
static	int		maxsequence = 0;	/* longest sequence defined */

        int     irsim_contline = 0;
static	sptr	xclock = NULL;		/* vectors which make up clock */
static	int		maxclock = 0;		/* longest clock sequence defined */

static	iptr	wlist = NULL;		/* list of nodes to be displayed */
static	iptr	wvlist = NULL;		/* list of vectors to be displayed */

static	CHAR	*filename;		/* current input file */
static	int		lineno = 0;		/* current line number */
static	int		column = 0;		/* current output column */
static	int		stoped_state = FALSE;	/* have we stoped ? */
					
static	Path	*cmdpath = NULL;	/* search path for cmd files */

static	TimeType	stepsize = 50000;	/* simulation step, in Delta's */
static	int		dcmdfile = 0;		/* display commands read from file */
static	int		ddisplay = 1;		/* if <>0 run "d" at end of step */

static	CHAR	*targv[MAXARGS];	/* array of tokens on command line */
static	int		targc;			/* number of args on command line */
static	CHAR	wildCard[MAXARGS];	/* set if corresponding arg has '*' */
static	CHAR	plus_minus[] = {x_("+")};	/* see apply() below */
static	CHAR	potchars[] = {x_("luxh.")};	/* set of potential characters */

static	CHAR	*first_file = NULL;	/* basename of network file read-in */

int    irsim_analyzerON = FALSE;	/* set when analyzer is running */
TimeType   irsim_sim_time0 = 0;		/* starting time (see flush_hist) */

#ifdef	POWER_EST
FILE  *irsim_caplogfile = NULL;	/* log file of cap transitions */
double irsim_toggled_cap = 0;	/* indicative of total power of chip */
float  irsim_vsupply = 5.0;		/* supply voltage for pwr estimation */
float  irsim_capstarttime = 0.0;
float  irsim_capstoptime = 0.0;
float  irsim_captime = 0.0;
float  irsim_powermult = 0.0;	/* to do power estimate in milliWatts */
int    irsim_pstep = 0;		/* Bool - end of step power display */
float  irsim_step_cap_x_trans = 0;	/* Stepwise C*trans count */
#endif

	/* forward references */
static int	expand(CHAR *string, CHAR **buffer, int *bufsize, CHAR wc);
static int  input(FILE *in);
static void	shift_args(int left);
static int  undefseq(nptr p, sptr *list, int *lmax);
static int  clockit(int n);

static	CHAR	not_in_stop[] = {x_("Can't do that while stoped, try \"C\"")};

void irsim_init_rsim(void)
{
	blist = NULL;
	slist = NULL;
	maxsequence = 0;
	irsim_contline = 0;
	xclock = NULL;
	maxclock = 0;
	wlist = NULL;
	wvlist = NULL;
	lineno = 0;
	column = 0;
	stoped_state = FALSE;
	cmdpath = NULL;
	stepsize = 50000;
	dcmdfile = 0;
	ddisplay = 1;
	first_file = NULL;
	irsim_analyzerON = FALSE;
	irsim_sim_time0 = 0;

#ifdef	POWER_EST
	irsim_caplogfile = NULL;
	irsim_toggled_cap = 0;
	irsim_vsupply = 5.0;
	irsim_capstarttime = 0.0;
	irsim_capstoptime = 0.0;
	irsim_captime = 0.0;
	irsim_powermult = 0.0;
	irsim_pstep = 0;
	irsim_step_cap_x_trans = 0;
#endif
}

/*
 * Routine to free all memory allocated in this module.
 */
void irsim_freersimmemory(void)
{
	Path *p;

	while (cmdpath != 0)
	{
		p = cmdpath;
		cmdpath = p->next;
		efree((CHAR *)p);
	}
}


/* 
 * Parse line into tokens, filling up targv and WildCard, and setting 'targc'
 */
static int parse_line(CHAR *line, int bufsize)
{
	CHAR           *extra;
	register int   i;
	register CHAR  ch;
	CHAR           wc;			/* wild card indicator */

	/* extra storage comes out of unused portion of line buffer */
	i = estrlen(line) + 1;
	bufsize -= i;
	extra = &line[i];
	targc = 0;
	while (i = *line++)
	{
		/* skip past white space */
		if (i <= ' ') continue;

		/* found start of new argument */
		if (targc == 0 && i == '|')
		{
			targc = 0;		/* comment line, stop now */
			return 0;
		}

		/* remember where argument begins */
		if (targc >= MAXARGS)
		{
			irsim_error(filename, lineno, x_("too many arguments in command"));
			targc = 0;
			return 0;
		}
		else
			targv[targc++] = --line;

		/* skip past text of argument, terminate with null character.
		 * While scanning argument remember if we see a "{" which marks
		 * the possible beginning of an iteration expression.
		 */
		wc = FALSE;
		i = 0;
		while ((ch = *line) > ' ')
		{
			if (ch == '*')
				wc = TRUE;
			else if (ch == ITERATOR_START)
				i = 1;
			line++;
		}
		*line++ = '\0';

		/* if the argument might contain one or more iterators, process
		 * it more carefully...
		 */
		if (i == 1)
		{
			if (expand(targv[--targc], &extra, &bufsize, wc))
			{
				targc = 0;
				return 0;
			}
		}
		else
			wildCard[targc - 1] = wc;
	}
	return 0;
}


/* 
 * Given a text string, expand any iterators it contains.  For example, the
 * string "out.{1:10}" expands into ten arguments "out.1", ..., "out.10".
 * The string can contain multiple iterators which will be expanded
 * independently, e.g., "out{1:10}{1:20:2}" expands into 100 arguments.
 * Buffer and bufsize describe a byte buffer which can be used for expansion.
 * Return 0 if expansion succeeds, 1 otherwise.
 */
static int expand(CHAR *string, CHAR **buffer, int *bufsize, CHAR wc)
{
	register CHAR  *p;
	CHAR           prefix[100], index[256];
	int            start, stop, nstep, length;

	/* copy string until we reach beginning of iterator */
	p = prefix;
	length = 0;
	while (*string)
	{
		if (*string == ITERATOR_START)
		{
			*p = 0;
			goto gotit;
		}
		*p++ = *string++;
	}
	*p = 0;

	/* if we get here, there was no iterator in the string, so save what
	 * we have as another argument.
	 */
	length = estrlen(prefix) + 1;
	if (length > *bufsize)
	{
		error(filename, lineno, x_("too many arguments in command"));
		return(1);
	}
	(void)estrcpy(*buffer, prefix);
	wildCard[targc] = wc;
	targv[targc++] = *buffer;
	*bufsize -= length;
	*buffer += length;
	return(0);

	/* gobble down iterator */
gotit:
	start = 0;
	stop = 0;
	nstep = 0;
	for(string += 1; *string >= '0' && *string <= '9'; string += 1)
		start = start * 10 + *string - '0';
	if (*string != ':')
		goto err;
	for(string += 1; *string >= '0' && *string <= '9'; string += 1)
		stop = stop * 10 + *string - '0';
	if (*string == ITERATOR_END)
		goto done;
	if (*string != ':')
		goto err;
	for(string += 1; *string >= '0' && *string <= '9'; string += 1)
		nstep = nstep * 10 + *string - '0';
	if (*string == ITERATOR_END)
		goto done;

err:
	irsim_error(filename, lineno, x_("syntax error in name iterator"));
	return(1);

done:	/* suffix starts just past '}' which terminates iterator */
	string += 1;

	/* figure out correct step size */
	if (nstep == 0)
		nstep = 1;
	else if (nstep < 0)
		nstep = -nstep;
	if (start > stop)
		nstep = -nstep;

	/* expand the iterator */
	while ((nstep > 0 && start <= stop) || (nstep < 0 && start >= stop))
	{
		(void)esnprintf(index, 256, x_("%s%d%s"), prefix, start, string);
		if (expand(index, buffer, bufsize, wc))
			return(1);
		start += nstep;
	}
	return(0);
}


static	int applyStart = 1;

/*
 * Apply given function to each argument on the command line.
 * Arguments are checked first to ensure they are the name of a node or
 * vector; wild-card patterns are allowed as names.
 * Either 'fun' or 'vfunc' is called with the node/vector as 1st argument:
 *	'fun' is called if name refers to a node.
 *	'vfun' is called if name refers to a vector.  If 'vfun' is NULL
 *	then 'fun' is called on each node of the vector.
 * The parameter (2nd argument) passed to the specified function will be:
 *	If 'arg' is the special constant 'plus_minus' then
 *	    if the name is preceded by a '-' pass a pointer to '-'
 *	    otherwise pass a pointer to '+'.
 *	else 'arg' is passed as is.
 */
static void apply(int (*fun)(), int (*vfun)(), CHAR  *arg)
{
	register CHAR  *p, *flag;
	register bptr  b;
	register int   i, j, found;

	for(i = applyStart ; i < targc; i += 1)
	{
		p = targv[i];
		if (arg == plus_minus)
		{
			if (*p == '-')
			{
				flag = p;
				p += 1;
			}
			else
				flag = plus_minus;
		} else
			flag = arg;

		found = 0;
		if (wildCard[i])
		{
			for(b = blist; b != NULL; b = b->next)
				if (irsim_str_match(p, b->name))
			{
				if (vfun != NULL)
					((int(*)(bptr,CHAR*))(*vfun))(b, flag);
				else
					for(j = 0; j < b->nbits; j += 1)
						((int(*)(nptr,CHAR*))(*fun))(b->nodes[j], flag);
				found = 1;
			}
			found += irsim_match_net(p, (int(*)(nptr, CHAR*))fun, flag);
		} else
		{
			nptr  n = irsim_find(p);

			if (n != NULL)
				found += ((int(*)(nptr,CHAR*))(*fun))(n, flag);
			else
			{
				for(b = blist; b != NULL; b = b->next)
					if (namesame(p, b->name) == 0)
				{
					if (vfun != NULL)
						((int(*)(bptr,CHAR*))(*vfun))(b, flag);
					else
						for(j = 0; j < b->nbits; j += 1)
							((int(*)(nptr,CHAR*))(*fun))(b->nodes[j], flag);
					found = 1;
					break;
				}
			}
		}
		if (found == 0)
			irsim_error(filename, lineno, x_("%s: No such node or vector"), p);
	}
}


/*
 * map a character into one of the potentials that a node can be set/compared
 */
static int ch2pot(CHAR ch)
{
	register int   i;
	register CHAR  *s;

	for(i = 0, s = x_("0ux1lUXhLUXH"); s[i] != '\0'; i++)
		if (s[i] == ch)
			return(i & (N_POTS - 1));

	irsim_error(filename, lineno, x_("%c: unknown node value"), ch);
	return(N_POTS);
}


/*
 * Open an input stream, and process commands from it.
 * Returns 0 if file could not be opened, 1 otherwise.
 */
static int finput(CHAR *name)
{
	FILE  *fp = NULL;
	int   olineno;
	CHAR  *ofname, *dummy;
	CHAR  pathname[256];

	if (*name == '/')				/* absolute path */
		fp = xopen(name, el_filetypetext, 0, &dummy);
	else
	{
		Path  *p;
		for(p = cmdpath; p != NULL; p = p->next)
		{
			(void)esnprintf(pathname, 256, x_("%s/%s"), p->name, name);
			fp = xopen(pathname, el_filetypetext, 0, &dummy);
			if (fp != NULL) break;
		}
	}
	if (fp == NULL)
		return(0);

	(void)estrcpy(pathname, name);
	ofname = filename;
	olineno = lineno;
	filename = pathname;
	lineno = 0;
	(void)input(fp);
	xclose(fp);
	filename = ofname;
	lineno = olineno;
	return(1);
}

#define	HashCmd(NM)	(((NM)[0] + (((NM)[1]) << 1)) & (CMDTBLSIZE - 1))

/*
 * Execute a builtin command or read commands from a '.cmd' file.
 */
static int exec_cmd(void)
{
	static CHAR       missing_args[] = {x_("missing arguments to \"%s\"")};
	static CHAR       many_args[] = {x_("too many arguments for \"%s\"")};
	register Command  *cmd;
	CHAR              *arg = targv[0];

	/* search command table, dispatch to handler, if any */
	for(cmd = cmdtbl[HashCmd(arg)]; cmd != NULL; cmd = cmd->next)
	{
		if (estrcmp(arg, cmd->name) == 0)
		{
			if (targc < cmd->nmin)
				irsim_error(filename, lineno, missing_args, cmd->name);
			else if (targc > cmd->nmax)
				irsim_error(filename, lineno, many_args, cmd->name);
			else
				return((*cmd->handler)());
			return(0);
		}
	}

	/* no built-in command found, try for a command file */
	if (targc == 1)
	{
		(void)estrcat(arg, x_(".cmd"));
		if (finput(arg))
			return(0);
		arg[estrlen(arg) - 4] = '\0';
	}

	irsim_error(filename, lineno, x_("unrecognized command: %s"), arg);
	return(0);
}

void irsim_issuecommand(CHAR *command, BOOLEAN showcommand)
{
	CHAR *line;
	REGISTER void *infstr;

	if (showcommand)
		ttyputmsg(x_("> %s"), command);
	infstr = initinfstr();
	formatinfstr(infstr, x_("%s\n"), command);
	line = returninfstr(infstr);
	parse_line(line, estrlen(line) + 1);
	if (targc > 0) (void)exec_cmd();
}

/* 
 * Read and execute commands read from input stream.
 */
static int input(FILE *in)
{
	static CHAR  line[LSIZE];		/* static: use only 1 line buffer */
	CHAR *str;
	int          ret;

	for(;;)
	{
		/* output prompt and flush output */
		if (in == stdin)
		{
			/* get command from messages window */
			str = ttygetlinemessages(x_("irsim> "));
			if (str == 0) break;
			estrcpy(line, str);
			estrcat(line, x_("\n"));
		} else
		{
			/* read command */
			if (xfgets(line, LSIZE, in)) break;
			if (dcmdfile)
				(void)ttyputmsg(x_("%s:%d> %s"), filename, lineno, line);
			estrcat(line, x_("\n"));
		}

		/* convert line into tokens */
		lineno += 1;
		lineno += irsim_contline;
		parse_line(line, LSIZE);
		if (targc != 0 && (ret = exec_cmd()) != 0)
			return(ret);

		/* if user typed ^C stop reading and return to top level */
		if (stopping(STOPREASONSIMULATE))
		{
			ttyputerr(x_("interrupt <%s @ %d>..."), filename, lineno);
			return(1);
		}
	}
	return(-1);
}


#define	CHECK_STOP() \
if (stoped_state) \
{ \
	irsim_error(filename, lineno, not_in_stop); \
	return(0); \
}

#define	UnAlias(N)	while ((N)->nflags & ALIAS) (N) = (N)->nlink


/*
 * Read and process a command file (@ command).
 */
static int cmdfile(void)
{
	if (! finput(targv[1]))
		irsim_error(filename, lineno, x_("cannot open %s for input"), targv[1]);

	return(0);
}


/*
 * Set or Print the search path for command files.
 */
static int docmdpath(void)
{
	Path  *p, **last;
	int   i;

	if (targc == 1)
	{
		/* echo current cmdpath */
		for(p = cmdpath; p != NULL; p = p->next)
			ttyputmsg(x_("%s "), p->name);
	} else
	{
		if (estrcmp(targv[1], x_("+")) == 0)
			shift_args(TRUE);
		else
		{
			while ((p = cmdpath) != NULL)
			{
				cmdpath = cmdpath->next;
				efree((CHAR *)p);
			}
		}
		last = &cmdpath;
		for(i = 1; i < targc; i++)
		{
			p = (Path *)emalloc(SIZEOF(Path) + estrlen(targv[i]), sim_tool->cluster);
			(void)estrcpy(p->name, targv[i]);
			p->next = NULL;
			*last = p;
			last = &p->next;
		}
	}
	return(0);
}


/*
 * Initialize the path to be the current working directory.
 */
void irsim_InitCmdPath(void)
{
	targc = 2;
	targv[1] = x_(".");
	(void)docmdpath();
}


/*
 * Set value of a node/vector to the requested value (hlux).
 */
static int setvalue(void)
{
	apply((int(*)())irsim_setin, 0, targv[0]);
	return(0);
}


/*
 * add/delete node to/from display list.
 */
static int xwatch(nptr n, CHAR *flag)
{
	UnAlias(n);

	if (! (n->nflags & MERGED))
	{
		if (*flag == '+')
			irsim_iinsert_once(n, &wlist);
		else
			irsim_idelete(n, &wlist);
	}
	return(1);
}


/*
 * add/delete vector to/from display list
 */
static int vwatch(bptr b, CHAR *flag)
{
	if (*flag == '+')
		irsim_iinsert_once((nptr) b, &wvlist);
	else
		irsim_idelete((nptr) b, &wvlist);

	return(1);
}


/* manipulate display list */
static int display(void)
{
	apply((int(*)())xwatch, (int(*)())vwatch, plus_minus);
	return(0);
}


/* display bit vector. */
static int dvec(bptr b)
{
	register int  i;
	CHAR          bits[250];

	i = estrlen(b->name) + 2 + b->nbits;
	if (column + i >= MAXCOL)
	{
		column = 0;
	}
	column += i;
	for(i = 0; i < b->nbits; i++)
		bits[i] = irsim_vchars[b->nodes[i]->npot];
	bits[i] = '\0';

	ttyputmsg(x_("%s=%s "), b->name, bits);

	return(1);
}


/* 
 * print value of specific node
 */
static void dnode(nptr n)
{
	register CHAR  *name;
	register int   i;

	name = pnode(n);
	UnAlias(n);
	i = estrlen(name) + ((n->nflags & MERGED) ? 23 : 3);
	if (column + i >= MAXCOL)
	{
		column = 0;
	}
	column += i;

	if (n->nflags & MERGED)
		ttyputmsg(x_("%s=<in transistor stack> "), name);
	else
		ttyputmsg(x_("%s=%c "), name, irsim_vchars[n->npot]);
}


/*
 * print current simulated time and the state of the event list.
 */
static void prtime(int col)
{
	if (col != 0)
		ttyputmsg(x_("time = %.2fns"), d2ns(irsim_cur_delta));
	if (irsim_npending)
		ttyputmsg(x_("; there are pending events (%d)"), irsim_npending);
}


/*
 * display node/vector values in display list
 */
static void pnwatchlist(void)
{
	register iptr  w;

	column = 0;

	/* print value of each watched bit vector. */
	for(w = wvlist; w != NULL; w = w->next)
		(void)dvec((bptr) w->inode);

	/* now print value of each watched node. */
	for(w = wlist; w != NULL; w = w->next)
		dnode((nptr) w->inode);

	prtime(column);
}


/*
 * Just append node to the list whose tail is pointed to by 'ptail'.
 */
static int get_nd_list(nptr n, nptr **ptail)
{
	if (! (n->nflags & VISITED))
	{
		n->nflags |= VISITED;
		n->n.next = NULL;
		**ptail = n;
		*ptail = &n->n.next;
	}
	return(1);
}


/*
 * display node values, either those specified or from display list
 */
static int pnlist(void)
{
	if (targc == 1)
		pnwatchlist();
	else
	{
		nptr  n = NULL, *ntail = &n;

		column = 0;

		/* first print any bit vectors the user has specified */
		apply((int(*)())get_nd_list, (int(*)())dvec, (CHAR *) &ntail);

		/* then print individual nodes collected in the list */
		for(; n != NULL; n->nflags &= ~VISITED, n = n->n.next)
			dnode(n);

		prtime(column);
	}
	return(0);
}


/*
 * set/clear trace bit in node
 */
static int xtrace(nptr n, CHAR *flag)
{
	UnAlias(n);

	if (n->nflags & MERGED)
	{
		ttyputmsg(x_("can't trace %s"), pnode(n));
		return(1);
	}

	if (*flag == '+')
		n->nflags |= WATCHED;
	else if (n->nflags & WATCHED)
	{
		ttyputmsg(x_("%s was watched; not any more"), pnode(n));
		n->nflags &= ~WATCHED;
	}

	return(1);
}


#ifdef POWER_EST
/*
 * set/clear powtrace bit in node
 */
static int xpowtrace(nptr n, CHAR *flag)
{
	UnAlias(n);

	if (n->nflags & MERGED)
	{
		ttyputmsg(x_("can't trace %s"), pnode(n));
		return(1);
	}

	if (*flag == '+')
		n->nflags |= POWWATCHED;
	else if (n->nflags & POWWATCHED)
	{
		ttyputmsg(x_("%s was capwatched; not any more"), pnode(n));
		n->nflags &= ~POWWATCHED;
	}

	return(1);
}
#endif

/*
 * set/clear trace bit in vector
 */
static int vtrace(bptr b, CHAR *flag)
{
	register int   i;

	if (*flag == '+')
		b->traced |= WATCHVECTOR;
	else
	{
		for(i = 0; i < b->nbits; i += 1)
			b->nodes[i]->nflags &= ~WATCHVECTOR;
		b->traced &= ~WATCHVECTOR;
	}
	return(1);
}

#ifdef POWER_EST
/*
 * set/clear powtrace bit in vector
 */
static int vpowtrace(bptr b, CHAR *flag)
{
	register int   i;

	if (*flag == '+')
		b->traced |= POWWATCHVECTOR;
	else
	{
		for(i = 0; i < b->nbits; i += 1)
			b->nodes[i]->nflags &= ~POWWATCHVECTOR;
		b->traced &= ~POWWATCHVECTOR;
	}
	return(1);
}
#endif

/*
 * just in case node appears in more than one bit vector, run through all
 * the vectors being traced and make sure the flag is set for each node.
 */
static void set_vec_nodes(int flag)
{
	register bptr  b;
	register int   i;

	for(b = blist; b != NULL; b = b->next)
		if (b->traced & flag)
			for(i = 0; i < b->nbits; i += 1)
				b->nodes[i]->nflags |= flag;
}


/*
 * set/clear stop bit in node
 */
static int nstop(nptr n, CHAR *flag)
{
	UnAlias(n);

	if (n->nflags & MERGED)
		return(1);

	if (*flag == '-')
		n->nflags &= ~STOPONCHANGE;
	else
		n->nflags |= STOPONCHANGE;

	return(1);
}


/*
 * set/clear stop bit in vector
 */
static int vstop(bptr b, CHAR *flag)
{
	register int   i;

	if (*flag == '+')
		b->traced |= STOPVECCHANGE;
	else
	{
		for(i = 0; i < b->nbits; i += 1)
			b->nodes[i]->nflags &= ~STOPVECCHANGE;
		b->traced &= ~STOPVECCHANGE;
	}
	return(1);
}


/*
 * mark nodes and vectors for tracing
 */
static int settrace(void)
{
	apply((int(*)())xtrace, (int(*)())vtrace, plus_minus);
	set_vec_nodes(WATCHVECTOR);
	return(0);
}

#ifdef POWER_EST
/*
 * mark nodes and vectors for cap tracing
 */
static int setpowtrace(void)
{
	apply((int(*)())xpowtrace, (int(*)())vpowtrace, plus_minus);
	set_vec_nodes(POWWATCHVECTOR);
	return(0);
}


/*
 * Helper routine for summing capacitance
 */
static int sumcapdoit(nptr n, float *capsum)
{
	CHAR  *nname = pnode(n);
	CHAR  *is_merge;

	UnAlias(n);

	if (! (n->nflags & (MERGED | ALIAS)))
		*capsum += n->ncap;

	return(0);
}


/*
 * Print sum of capacitance of nodes
 */
static int sumcap(void)
{
	float capsum = 0;

	ttyputmsg(x_("Sum of nodal capacitances: "));
	irsim_walk_net1(sumcapdoit, (CHAR *)&capsum);
	ttyputmsg(x_("%f pF"), capsum);
	return(0);
}


/*
 * Set the supply voltage to a known value -- used in calculating power
 */
static int setvsupply(void)
{
	if (targc == 2)
		irsim_vsupply = eatof(targv[1]);
	ttyputmsg(x_("Supply Voltage = %4.2f Volts"), irsim_vsupply);
	return(0);
}


/*
 * Toggle display of power estimation at end of each step
 */
static int togglepstep(void)
{
	irsim_pstep = !irsim_pstep;
	if (irsim_pstep)
		ttyputmsg(x_("Power display enabled"));
	else
		ttyputmsg(x_("Power display disbled"));
	return(0);
}
#endif



/*
 * mark nodes and vectors for stoping
 */
static int setstop(void)
{
	apply((int(*)())nstop, (int(*)())vstop, plus_minus);
	set_vec_nodes(STOPVECCHANGE);
	return(0);
}


/*
 * define bit vector
 */
static int dovector(void)
{
	register nptr  n;
	register bptr  b, last;
	register int   i;

	if (irsim_find(targv[1]) != NULL)
	{
		irsim_error(filename, lineno, x_("'%s' is a node, can't be a vector"),
			targv[1]);
		return(0);
	}

	/* get rid of any vector with the same name */
	for(b = blist, last = NULL; b != NULL; last = b, b = b->next)
		if (estrcmp(b->name, targv[1]) == 0)
	{
		if (undefseq((nptr) b, &slist, &maxsequence) ||
			undefseq((nptr) b, &xclock, &maxclock))
		{
			irsim_error(filename, lineno,
				x_("%s is a clock/sequence; can't change it while stoped"),
					b->name);
			return(0);
		}
		irsim_idelete((nptr) b, &wvlist);	/* untrace its nodes */
		if (last == NULL)
			blist = b->next;
		else
			last->next = b->next;		/* remove from display list */
		(void)vtrace(b, x_("-"));
		if (irsim_analyzerON)
			irsim_RemoveVector(b);
		efree(b->name);
		efree((CHAR *)b);
		break;
	}
	b = (bptr)emalloc(SIZEOF(Bits) + (targc - 3) * SIZEOF(nptr), sim_tool->cluster);
	if (b == NULL || (b->name = (CHAR *)emalloc((estrlen(targv[1]) + 1) * SIZEOFCHAR, sim_tool->cluster)) == NULL)
	{
		if (b) efree((CHAR *)b);
			irsim_error(filename, lineno, x_("Not enough memory for vector"));
		return(0);
	}
	b->traced = 0;
	b->nbits = 0;
	(void)estrcpy(b->name, targv[1]);

	for(i = 2; i < targc; i += 1)
	{
		if ((n = irsim_find(targv[i])) == NULL)
			irsim_error(filename, lineno, x_("cannot find node %s"), targv[i]);
		else
		{
			UnAlias(n);
			if (n->nflags & MERGED)
				irsim_error(filename, lineno, x_("%s can not be part of a vector"),
					pnode(n));
			else
				b->nodes[b->nbits++] = n;
		}
	}

	if (b->nbits == targc - 2)
	{
		b->next = blist;
		blist = b;
	} else
	{
		efree(b->name);
		efree((CHAR *)b);
	}

	return(0);
}


/* set bit vector */
static int setvector(void)
{
	register bptr  b;
	register int   i;
	CHAR           *val = targv[2];

	/* find vector */
	for(b = blist; b != NULL; b = b->next)
	{
		if (namesame(b->name, targv[1]) == 0)
			goto got_it;
	}

	irsim_error(filename, lineno, x_("%s: No such vector"), targv[1]);
	return(0);

got_it:
	/* set nodes */
	if ((int)estrlen(targv[2]) != b->nbits)
	{
		irsim_error(filename, lineno, x_("wrong number of bits for this vector"));
		return(0);
	}
	for(i = 0; i < b->nbits; i++)
	{
		if ((val[i] = potchars[ch2pot(val[i])]) == '.')
			return(0);
	}
	for(i = 0; i < b->nbits; i++)
		(void)irsim_setin(b->nodes[i], val++);
	return(0);
}


static int CompareVector(nptr *np, CHAR *name, int nbits, CHAR *mask, CHAR *value)
{
	int   i, val;
	nptr  n;

	if ((int)estrlen(value) != nbits)
	{
		irsim_error(filename, lineno, x_("wrong number of bits for value"));
		return 0;
	}
	if (mask != NULL && (int)estrlen(mask) != nbits)
	{
		irsim_error(filename, lineno, x_("wrong number of bits for mask"));
		return 0;
	}

	for(i = 0; i < nbits; i++)
	{
		if (mask != NULL && mask[i] != '0')
			continue;
		n = np[i];
		if ((val = ch2pot(value[i])) >= N_POTS)
			return 0;
		if (val == X_X) val = X;
			if (n->npot != val)
				goto fail;
	}
	return 0;

fail:
	return 1;
}


typedef struct
{
	nptr  node;
	bptr  vec;
	int   num;
} Find1Arg;


static int SetNode(nptr nd, Find1Arg *find_one)
{
	find_one->node = nd;
	find_one->num++;
	return(1);
}

static int SetVector(bptr bp, Find1Arg *find_one)
{
	find_one->vec = bp;
	find_one->num++;
	return(1);
}


/* find vector or node (1st argument) */
static void FindOne(Find1Arg *f)
{
	targc = 2;
	f->num = 0;
	f->vec = NULL;
	f->node = NULL;
	apply((int(*)())SetNode, (int(*)())SetVector, (CHAR *) f);
}
  

/* assert a bit vector */
static int doAssert(void)
{
	CHAR      *mask, *value, *name;
	Find1Arg  f;
	int i, nbits, comp = 0;
	nptr *nodes;
	REGISTER void *infstr;

	if (targc == 4)
	{
		mask = targv[2];
		value = targv[3];
	} else
	{
		mask = NULL;
		value = targv[2];
	}

	FindOne(&f);

	if (f.num > 1)
		irsim_error(filename, lineno, x_("%s matches more than one node or vector"),
			targv[1]);
	else if (f.node != NULL)
	{
		name = pnode(f.node);
		UnAlias(f.node);
		comp = CompareVector(&f.node, name, 1, mask, value);
		nodes = &f.node;
		nbits = 1;
	}
	else if (f.vec != NULL)
	{ 
		comp = CompareVector(f.vec->nodes, f.vec->name, f.vec->nbits, mask, value);
		name = f.vec->name;
		nbits = f.vec->nbits;
		nodes = f.vec->nodes; 
	}
	if (comp != 0)
	{
		ttyputmsg(x_("(%s, %d): assertion failed on '%s' "),
			filename, lineno, name);
		infstr = initinfstr();
		for(i = 0; i < nbits; i++)
		{
			if (mask != NULL && mask[i] != '0')
			{
				addtoinfstr(infstr, '-');
				value[i] = '-';
			}
			else
				addtoinfstr(infstr, irsim_vchars[nodes[i]->npot]);
		}
		ttyputmsg(x_("Want (%s) but got (%s)"), value, returninfstr(infstr));
	}
	return(0);
}



static int doUntil(void)
{
	CHAR      *mask, *value, *name;
	Find1Arg  f;
	int i, nbits, ccount, cnt = 0, comp = 0;
	nptr *nodes;
	REGISTER void *infstr;

	if (targc == 5)
	{
		mask = targv[2];
		value = targv[3];
		ccount = eatoi(targv[4]);
	} else
	{
		mask = NULL;
		value = targv[2];
		ccount = eatoi(targv[3]);
	}

	FindOne(&f);

	if (f.num > 1)
		irsim_error(filename, lineno, x_("%s matches more than one node or vector"),
			targv[1]);
	else if (f.node != NULL)
	{
		name = pnode(f.node);
		UnAlias(f.node);
		targc = 1;
		while ((cnt <= ccount) && (comp = CompareVector(&f.node, name, 1, mask, value) != 0))
		{
			cnt++;
			clockit(1);
		}
		nodes = &f.node;
		nbits = 1;
	}
	else if (f.vec != NULL)
	{
		targc = 1;
		while ((cnt <= ccount) && (comp = CompareVector(f.vec->nodes, f.vec->name, f.vec->nbits, mask,
			value) != 0))
		{
			cnt++;
			clockit(1);
		}
		name = f.vec->name;
		nbits = f.vec->nbits;
		nodes = f.vec->nodes;
	}
	if (comp != 0)
	{
		ttyputmsg(x_("(%s, %d): assertion failed on '%s' "),
			filename, lineno, name);
		infstr = initinfstr();
		for(i = 0; i < nbits; i++)
		{
			if (mask != NULL && mask[i] != '0')
			{
				addtoinfstr(infstr, '-');
				value[i] = '-';
			}
			else
				addtoinfstr(infstr, irsim_vchars[nodes[i]->npot]);
		}
		ttyputmsg(x_("Want (%s) but got (%s)"), value, returninfstr(infstr));
	}
	return(0);
}


static	nptr aw_trig ; /* keeps current assertWhen trigger */
static awptr aw_p ;   /* track pointer on the current assertWhen list */

static int setupAssertWhen(nptr n, CHAR *val)
{
	register awptr p;

	p = (awptr)emalloc(sizeof(assertWhen), sim_tool->cluster);
	p->node = n ;
	p->val  = *val ;
	p->nxt = NULL ;

	if (aw_trig->awpending == NULL)
	{
		/* first time */
		aw_trig->awpending = p ;
		aw_p = p ;
	} else
	{
		/* more than 1 matching nodes */
		aw_p->nxt = p ;
		aw_p = p ;
	}
	return 1;
}

static int doAssertWhen(void)
{
	Find1Arg  trig;

	FindOne(&trig);

	if (trig.num > 1)
		irsim_error(filename, lineno, x_("%s matches more than one node or vector"),
			targv[1]);
	else if (trig.node != NULL)
	{
		applyStart = 3; targc  = 4;

		UnAlias(trig.node);
		aw_trig = trig.node ;
		aw_trig->awpot = ch2pot(*targv[2]);
		apply((int(*)())setupAssertWhen, 0, targv[4]);

		applyStart = 1; targc = 4;
	}
	else if (trig.vec != NULL) 
		irsim_error(filename, lineno, x_("trigger to assertWhen %s can't be a vector"),
			targv[1]);
	return(0);
}

void	irsim_evalAssertWhen(nptr n)
{
	register awptr p, p2 ; 
	CHAR cval[2] ; 
	CHAR *name, comp;

	cval[0] = 0 ;
	cval[1] = 0 ;
	for (p = n->awpending ; p ; )
	{
		cval[0] = p->val ; 
		name = pnode(p->node);
		comp = CompareVector(&(p->node), name, 1, NULL, cval);
		if (comp != 0) 
			ttyputmsg(x_("(%s, %d): assertion failed on '%s' "),
				filename, lineno, name);
		p2 = p ; p = p->nxt ;
		efree((CHAR *)p2);
	}
	n->awpending = NULL ;
}




static int collect_inputs(nptr n, nptr inps[])
{
	if ((n->nflags & (INPUT|ALIAS|POWER_RAIL|VISITED|INPUT_MASK)) == INPUT)
	{
		n->n.next = inps[n->npot];
		inps[n->npot] = n;
		n->nflags |= VISITED;
	}
	return(0);
}


/* display current inputs */
static int inputs(void)
{
	register iptr  list;
	register nptr  n;
	nptr           inptbl[N_POTS];

	inptbl[HIGH] = inptbl[LOW] = inptbl[X] = NULL;
	irsim_walk_net1((int(*)(nptr, CHAR*))collect_inputs, (CHAR *)inptbl);

	ttyputmsg(x_("h inputs: "));
	for(list = irsim_hinputs; list != NULL; list = list->next)
		ttyputmsg(x_("%s "), pnode(list->inode));
	for(n = inptbl[HIGH]; n != NULL; n->nflags &= ~VISITED, n = n->n.next)
		ttyputmsg(x_("%s "), pnode(n));
	ttyputmsg(x_("l inputs: "));
	for(list = irsim_linputs; list != NULL; list = list->next)
		ttyputmsg(x_("%s "), pnode(list->inode));
	for(n = inptbl[LOW]; n != NULL; n->nflags &= ~VISITED, n = n->n.next)
		ttyputmsg(x_("%s "), pnode(n));
	ttyputmsg(x_("u inputs: "));
	for(list = irsim_uinputs; list != NULL; list = list->next)
		ttyputmsg(x_("%s "), pnode(list->inode));
	for(n = inptbl[X]; n != NULL; n->nflags &= ~VISITED, n = n->n.next)
		ttyputmsg(x_("%s "), pnode(n));
	return(0);
}


/* set stepsize */
static int setstep(void)
{
	if (targc == 1)
		ttyputmsg(x_("stepsize = %f"), d2ns(stepsize));
	else if (targc == 2)
	{
		TimeType  newsize = ns2d(eatof(targv[1]));

		if (newsize <= 0)
			irsim_error(filename, lineno, x_("bad step size: %s"), targv[1]);
		else
			stepsize = newsize;
	}
	return(0);
}


/*
 * Display traced vectors that just changed.  There should be at least one.
 */
void irsim_disp_watch_vec(long which)
{
	register bptr  b;
	register int   i;
	CHAR           temp[20];

	which &= (WATCHVECTOR | STOPVECCHANGE);
	(void)esnprintf(temp, 20, x_(" @ %.2fns "), d2ns(irsim_cur_delta));
	ttyputmsg(x_("%s"), temp);
	column = estrlen(temp);
	for(b = blist; b != NULL; b = b->next)
	{
		if ((b->traced & which) == 0)
			continue;
		for(i = b->nbits - 1; i >= 0; i--)
			if ((long)b->nodes[i]->c.time == irsim_cur_delta)
				break;
		if (i >= 0)
			(void)dvec(b);
	}
}


static void EnterStopState(void)
{
	CHAR  *ofname = filename;
	int   olineno = lineno;

	ttyputmsg(x_("--> STOP  "));
	prtime(0);
	filename = x_("STOP");
	lineno = 0;
	stoped_state = TRUE;
	while (input(stdin) <= 0);	/* ignore quit */
	stoped_state = FALSE;
	lineno = olineno;
	filename = ofname;
}


/* 
 * Settle network until the specified stop time is reached.
 * Premature returns (before stop time) indicate that a node/vector whose
 * stop-bit set has just changed value, so popup a stdin command interpreter.
 */
static TimeType relax(TimeType stoptime)
{
	while (irsim_step(stoptime))
	{
		EnterStopState();
	}
	return(irsim_cur_delta - stoptime);
}


/*
 * relax network, optionally set stepsize
 */
static int dostep(void)
{
	TimeType newsize;
#ifdef POWER_EST
	TimeType pstepstart;
#endif
	CHECK_STOP();

	if (targc == 2)
	{
		newsize = ns2d(eatof(targv[1]));
		if (newsize <= 0)
		{
			irsim_error(filename, lineno, x_("bad step size: %s"), targv[1]);
			return(0);
		}
	}
	else
		newsize = stepsize;

#ifdef POWER_EST
	pstepstart = irsim_cur_delta;
	irsim_step_cap_x_trans = 0;
#endif

	(void)relax(irsim_cur_delta + newsize);
	if (ddisplay)
		pnwatchlist();

#ifdef POWER_EST
	if (irsim_pstep)
		ttyputmsg(x_("Dynamic power estimate for powtrace'd nodes on last step = %f mW"),
			irsim_step_cap_x_trans*irsim_vsupply*irsim_vsupply/(2*(d2ns(irsim_cur_delta-pstepstart))));
#endif

	return(0);
}


/*
 * display info about a node
 */
static int quest(void)
{
	apply((int(*)())irsim_info, 0, targv[0]);
	return(0);
}


/*
 * exit this level of command processing
 */
static int quit(void)
{
	return(-1);
}


/*
 * return to command level
 */
static int doexit(void)
{
	irsim_TerminateAnalyzer();

	/* exit((targc == 2) ? atoi(targv[1]) : 0); */
	return(-1);
}


/*
 * destroy sequence for given node/vector: update sequence list and length.
 * return -1 if we can't destroy the sequence (in stopped state).
 */
static int undefseq(nptr p, sptr *list, int *lmax)
{
	register sptr  u, t;
	register int   i;

	for(u = NULL, t = *list; t != NULL; u = t, t = t->next)
		if (t->ptr.n == p)
			break;
	if (t)
	{
		if (stoped_state)	/* disallow changing sequences if stoped */
			return(-1);
		if (u == NULL)
			*list = t->next;
		else
			u->next = t->next;
		efree((CHAR *)t);
		for(i = 0, t = *list; t != NULL; t = t->next)
			if (t->nvalues > i) i = t->nvalues;
				*lmax = i;
	}
	return(0);
}


/*
 * process command line to yield a sequence structure.  first arg is the
 * name of the node/vector for which the sequence is to be defined, second
 * and following args are the values.
 */
static void defsequence(sptr *list, int *lmax)
{
	register sptr  s;
	register nptr  n;
	register bptr  b;
	register int   i;
	register CHAR  *q;
	int            which, size;

	/* if no arguments, get rid of all the sequences we have defined */
	if (targc == 1)
	{
		while (*list != NULL)
			(void)undefseq((*list)->ptr.n, list, lmax);
		return;
	}

	/* see if we can determine if name is for node or vector */
	for(b = blist; b != NULL; b = b->next)
		if (namesame(b->name, targv[1]) == 0)
	{
		which = 1;    size = b->nbits;    goto okay;
	}

	n = irsim_find(targv[1]);
	if (n == NULL)
	{
		irsim_error(filename, lineno, x_("%s: No such node or vector"), targv[1]);
		return;
	}
	UnAlias(n);
	if (n->nflags & MERGED)
	{
		irsim_error(filename, lineno, x_("%s can't be part of a sequence"), pnode(n));
		return;
	}
	which = 0; size = 1;

okay:
	if (targc == 2)	/* just destroy the given sequence */
	{
		(void)undefseq(which ? (nptr) b : n, list, lmax);
		return;
	}

	/* make sure each value specification is the right length */
	for(i = 2; i < targc; i += 1)
		if ((int)estrlen(targv[i]) != size)
	{
		irsim_error(filename, lineno,
			x_("value \"%s\" is not compatible with size of %s (%d)"),
				targv[i], targv[2], size);
		return;
	}

	s = (sptr)emalloc(SIZEOF(sequence) + size * (targc - 2) - 1, sim_tool->cluster);
	if (s == NULL)
	{
		irsim_error(filename, lineno, x_("Insufficient memory for sequence"));
		return;
	}
	s->which = which;
	s->vsize = size;
	s->nvalues = targc - 2;
	if (which)	s->ptr.b = b;
		else	s->ptr.n = n;

	/* process each value specification saving results in sequence */
	for(q = s->values, i = 2; i < targc; i += 1)
	{
		register CHAR  *p;

		for(p = targv[i]; *p != 0; p++)
			if ((*q++ = potchars[ch2pot(*p)]) == '.')
		{
			efree((CHAR *)s);
			return;
		}
	}

	/* all done!  remove any old sequences for this node or vector. */
	(void)undefseq(s->ptr.n, list, lmax);

	/* insert result onto list */
	s->next = *list;
	*list = s;
	if (s->nvalues > *lmax)
		*lmax = s->nvalues;
}


/*
 * mark any vector that contains a deleted node (used by netupdate).
 */
static int mark_deleted_vectors(void)
{
	register bptr  b;
	register int   i, total = 0;

	for(b = blist; b != NULL; b = b->next)
	{
		for(i = b->nbits - 1; i >= 0; i--)
		{
			if (b->nodes[i]->nflags & DELETED)
			{
				b->traced = DELETED;
				total ++;
				break;
			}
			UnAlias(b->nodes[i]);
		}
	}
	return(total);
}


/*
 * Remove all deleted nodes/vectors from the sequence list
 */
static int rm_from_seq(sptr *list)
{
	register sptr  s;
	register int   max;

	max = 0;
	while ((s = *list) != NULL)
	{
		if (((s->which) ? s->ptr.b->traced : s->ptr.n->nflags) & DELETED)
		{
			*list = s->next;
			efree((CHAR *)s);
		} else
		{
			if (s->which == 0)
				UnAlias(s->ptr.n);

			if (s->nvalues > max)
				max = s->nvalues;
			list = &(s->next);
		}
	}
	return(max);
}


/*
 * Remove any deleted node/vector from any lists in which it may appear.
 */
void irsim_rm_del_from_lists(void)
{
	register iptr  w, *list;
	int            vec_del;

	vec_del = mark_deleted_vectors();

	maxsequence = rm_from_seq(&slist);
	maxclock = rm_from_seq(&xclock);

	if (irsim_analyzerON)
		irsim_RemoveAllDeleted();

	for(list = &wvlist; (w = *list) != NULL;)
	{
		if (((bptr) (w->inode))->traced & DELETED)
		{
			*list = w->next;
			FreeInput(w);
		}
		else
			list = &w->next;
	}

	for(list = &wlist; (w = *list) != NULL;)
	{
		if (w->inode->nflags & DELETED)
		{
			*list = w->next;
			FreeInput(w);
		} else
		{
			UnAlias(w->inode);
			list = &w->next;
		}
	}

	if (vec_del)
	{
		register bptr  b, *lst;

		for(lst = &blist; (b = *lst) != NULL;)
		{
			if (b->traced & DELETED)
			{
				*lst = b->next;
				efree(b->name);
				efree((CHAR *)b);
			}
			else
				lst = &b->next;
		}
	}
}


/*
 * set each node/vector in sequence list to its next value
 */
static void vecvalue(sptr list, int index)
{
	register int   offset, i;
	register nptr  *n;

	for(; list != NULL; list = list->next)
	{
		offset = list->vsize * (index % list->nvalues);
		n = (list->which == 0) ? &list->ptr.n : list->ptr.b->nodes;
		for(i = 0; i < list->vsize; i++)
			(void)irsim_setin(*n++, &list->values[offset++]);
	}
}


/*
 * setup sequence of values for a node
 */
static int setseq(void)
{
	CHECK_STOP();
	/* process sequence and add to list */
	defsequence(&slist, &maxsequence);
	return(0);
}


/*
 * define clock sequences(s)
 */
static int setclock(void)
{
	CHECK_STOP();
	/* process sequence and add to clock list */
	defsequence(&xclock, &maxclock);
	return(0);
}


/*
 * Step each clock node through one simulation step
 */
static int step_phase(void)
{
	static int  which_phase = 0;

	vecvalue(xclock, which_phase);
	if (relax(irsim_cur_delta + stepsize))
		return(1);
	which_phase = (which_phase + 1) % maxclock;
	return(0);
}


/* Do one simulation step */
static int dophase(void)
{
	CHECK_STOP();

	(void)step_phase();
	if (ddisplay)
		pnwatchlist();

	return(0);
}


/*
 * clock circuit specified number of times
 */
static int clockit(int n)
{
	register int  i;

	if (xclock == NULL)
		irsim_error(filename, lineno, x_("no clock nodes defined!"));
	else
	{
		/* run 'em by setting each clock node to successive values of its
		 * associated sequence until all phases have been run.
		 */
		while (n-- > 0)
			for(i = 0; i < maxclock; i += 1)
				if (step_phase())
					goto done;

		/* finally display results if requested to do so */
done:
		if (ddisplay)
			pnwatchlist();
	}
	return(maxclock - i);
}


/*
 * clock circuit through all the input vectors previously set up
 */
static int runseq(void)
{
	register int  i, n = 1;

	CHECK_STOP();

	/* calculate how many clock cycles to run */
	if (targc == 2)
	{
		n = eatoi(targv[1]);
		if (n <= 0)
			n = 1;
	}

	/* run 'em by setting each input node to successive values of its
	 * associated sequence.
	 */
	if (slist == NULL)
		irsim_error(filename, lineno, x_("no input vectors defined!"));
	else
		while (n-- > 0)
			for(i = 0; i < maxsequence; i += 1)
	{
		vecvalue(slist, i);
		if (clockit(1))
			return(0);
		if (ddisplay)
			pnwatchlist();
	}

	return(0);
}


/*
 * process "c" command line
 */
static int doclock(void)
{
	register int  n = 1;

	if (stoped_state)		/* continue after stop */
		return(1);

	/* calculate how many clock cycles to run */
	if (targc == 2)
	{
		n = eatoi(targv[1]);
		if (n <= 0)
			n = 1;
	}

	(void)clockit(n);		/* do the hard work */
	return(0);
}


/*
 * output message to console/log file
 */
static int domsg(void)
{
	int  n;
	void *infstr;

	infstr = initinfstr();
	for(n=1; n<targc; n++)
	{
		if (n != 1) addstringtoinfstr(infstr, x_(" "));
		addstringtoinfstr(infstr, targv[n]);
	}
	ttyputmsg(x_("%s "), returninfstr(infstr));
	return(0);
}


/*
 * Return a number whose bits corresponding to the index of 'words' match
 * the argument.
 * If 'offwrd' is given then that argument turns all bits off.
 * If 'offwrd' is given and the argument is '*' turns all bits on.
 * if the argument is '?' the display the avaliable options (words).
 * With no arguments just prints the word whose corresponding bit is set.
 */
static int do_flags(int bits, CHAR *name, CHAR *offwrd, CHAR *words[])
{
	int  i, t, tmp;

	if (targc == 1)
	{
		ttyputmsg(x_("%s: "), name);
		if (bits == 0 && offwrd != NULL)
			ttyputmsg(offwrd);
		else
		{
			for(i = 0; words[i] != NULL; i++)
				if (bits & (1 << i))
					ttyputmsg(x_(" %s"), words[i]);
		}
	}
	else if (targc == 2 && estrcmp(targv[1], x_("?")) == 0)
	{
		ttyputmsg(x_("%s options are:"), name);
		if (offwrd)
			ttyputmsg(x_("[*][%s]"), offwrd);
		for(t = '[', i = 0; words[i] != NULL; i++, t = ' ')
			ttyputmsg(x_("%c%s"), t, words[i]);
		ttyputmsg(x_("]"));
	}
	else if (targc == 2 && offwrd != 0 && estrcmp(targv[1], offwrd) == 0)
	{
		bits = 0;
	}
	else if (targc == 2 && offwrd != 0 && namesame(targv[1], x_("*")) == 0)
	{
		for(i = 0; words[i] != NULL; i++);
			bits = (1 << i) - 1;
	}
	else
	{
		for(t = 1, tmp = bits, bits = 0; t < targc; t++)
		{
			for(i = 0; words[i] != NULL; i++)
				if (namesame(words[i], targv[t]) == 0)
			{
				bits |= (1 << i);
				break;
			}
			if (words[i] == NULL)
			{
				irsim_error(filename, lineno, x_("%s: Invalid %s option"),
					targv[t], name);
				bits = tmp;
				break;
			}
		}
	}
	return(bits);
}


/* set debugging level */
static int setdbg(void)
{
	static CHAR  *dbg[] = { x_("ev"), x_("dc"), x_("tau"), x_("taup"), x_("spk"), x_("tw"), NULL };
	int          newdbg;

	newdbg = do_flags(irsim_debug, x_("Debug"), x_("off"), dbg);
	if (newdbg != irsim_debug)
	{
		irsim_debug = newdbg;
		targc = 1;
		(void)do_flags(irsim_debug, x_("Debug is now"), x_("OFF"), dbg);
	}
	return(0);
}


/*
 * set irsim_treport parameter
 */
static int setreport(void)
{
	static CHAR  *rep[] = { x_("decay"), x_("delay"), x_("tau"), x_("tcoord"), NULL };

	irsim_treport = do_flags(irsim_treport, x_("report"), x_("none"), rep);
	return(0);
}


/*
 * set which evaluation model to use
 */
static int setmodel(void)
{
	static CHAR  *m_name[] = { x_("linear"), x_("switch"), NULL };
	int          newmod;

	/* note: the following will only work for 2 models! */
	newmod = do_flags(irsim_model_num + 1, x_("model"), NULL, m_name) - 1;

	if (newmod != irsim_model_num)
	{
		irsim_model_num = newmod;
		irsim_model = irsim_model_table[irsim_model_num];
		irsim_NewModel(newmod);
	}
	return(0);
}


/*
 * set up or finish off logfile
 */
static int setlog(void)
{
	ttyputerr(x_("No logging facility...use the one in Electric"));
	return(0);
}


#ifdef POWER_EST
/*
 * Helper routine for printing capacitance and power totals.
 */
static int capsummer(nptr n)
{
	UnAlias(n);

	if ((! (n->nflags & (MERGED | ALIAS))) && (n->nflags & POWWATCHED))
	{
		ttyputmsg(x_(" %-35s\t%.3f\t%5d\t%f\t%f"),
			pnode(n), n->ncap, n->toggles,
			n->toggles*n->ncap*irsim_powermult,
			n->toggles*n->ncap / irsim_toggled_cap);
	}
	return(0);
}

/*
 * set up or finish off logfile
 */
static int setcaplog(void)
{
	if (irsim_caplogfile != NULL)
	{
		(void)fclose(irsim_caplogfile);
		irsim_caplogfile = NULL;
		irsim_capstoptime = d2ns(irsim_cur_delta);
		irsim_captime = irsim_capstoptime - irsim_capstarttime;
		irsim_powermult = irsim_vsupply*irsim_vsupply/(2*irsim_captime);
		irsim_walk_net(capsummer);
		ttyputmsg(x_("Dynamic power estimate for powtrace'd nodes = %f Watts (%f)"),
			irsim_toggled_cap * irsim_powermult * 1e-3, irsim_toggled_cap);
	}

	if (targc == 2)
	{
		CHAR  *mode = x_("w");
		CHAR  *s = targv[1];

		if (*s == '+')
		{
			s++;
			mode = x_("a");
		}
		if ((caplogfile = efopen(s, mode)) == NULL)
		irsim_error(filename, lineno, x_("cannot open log file %s for output"), s);
		irsim_capstarttime = d2ns(irsim_cur_delta);
	}
	return(0);
}
#endif

/*
 * restore state of network
 */
static int do_rdstate(void)
{
	CHAR  *err;

	CHECK_STOP();

	if (err = irsim_rd_state(targv[1], (targv[0][1] == '<') ? TRUE : FALSE))
		irsim_error(filename, lineno, err);
	return(0);
}


/*
 * restore state of network
 */
static int do_wrstate(void)
{
	if (irsim_wr_state(targv[1]))
		irsim_error(filename, lineno, x_("can not write state file: %s"), targv[1]);
	return(0);
}


/*
 * set decay parameter
 */
static int setdecay(void)
{
	if (targc == 1)
	{
		if (irsim_tdecay == 0)
			ttyputmsg(x_("decay = No decay"));
		else
			ttyputmsg(x_("decay = %.2fns"), d2ns(irsim_tdecay));
	} else
	{
		irsim_tdecay = ns2d(eatof(targv[1]));
		if (irsim_tdecay < 0)
			irsim_tdecay = 0;
	}
	return(0);
}


/*
 * set unitdelay parameter
 */
static int setunit(void)
{
	if (targc == 1)
	{
		if (irsim_tunitdelay == 0)
			ttyputmsg(x_("unitdelay = OFF"));
		else
			ttyputmsg(x_("unitdelay = %.2f"), d2ns(irsim_tunitdelay));
	} else
	{
		if ((irsim_tunitdelay = (int) ns2d(eatof(targv[1]))) < 0)
			irsim_tunitdelay = 0;
	}
	return(0);
}


/*
 * print traceback of node's activity and that of its ancestors
 */
static void cpath(nptr n, int level)
{
	static TimeType  ptime;		/* previous time, used during trace back */

	/* no last transition! */
	if ((n->nflags & MERGED) || n->t.cause == NULL)
	{
		ttyputmsg(x_("  there is no previous transition!"));
	}
	else if (n->t.cause == irsim_inc_cause)
	{
		if (level == 0)
			ttyputmsg(x_("  previous transition due to incremental update"));
		else
			ttyputmsg(x_("  transition of %s due to incremental update"), pnode(n));
	}

	/* here if we come across a node which has changed more recently than
	 * the time reached during the backtrace.  We can't continue the
	 * backtrace in any reasonable fashion, so we stop here.
	 */
	else if (level != 0 && n->c.time > ptime)
	{
		ttyputmsg(x_("  transition of %s, which has since changed again"), pnode(n));
	}
	/* here if there seems to be a cause for this node's transition.
	 * If the node appears to have 'caused' its own transition (n->t.cause
	 * == n), that means it was input.  Otherwise continue backtrace...
	 */
	else if (n->t.cause == n)
	{
		ttyputmsg(x_("  %s -> %c @ %.2fns , node was an input"),
			pnode(n), irsim_vchars[n->npot], d2ns(n->c.time));
	}
	else if (n->t.cause->nflags & VISITED)
	{
		ttyputmsg(x_("  ... loop in traceback"));
	}
	else
	{
		TimeType  delta_t = n->c.time - n->t.cause->c.time;

		n->nflags |= VISITED;
		ptime = n->c.time;
		cpath(n->t.cause, level + 1);
		n->nflags &= ~VISITED;
		if (delta_t < 0)
			ttyputmsg(x_("  %s -> %c @ %.2fns   (??)"), pnode(n),
				irsim_vchars[n->npot], d2ns(n->c.time));
		else
			ttyputmsg(x_("  %s -> %c @ %.2fns   (%.2fns)"), pnode(n),
				irsim_vchars[n->npot], d2ns(n->c.time), d2ns(delta_t));
	}
}


static int do_cpath(nptr n)
{
	ttyputmsg(x_("critical path for last transition of %s:"), pnode(n));
	UnAlias(n);
	cpath(n, 0);
	return(1);
}


/*
 * discover and print critical path for node's last transistion
 */
static int dopath(void)
{
	apply((int(*)())do_cpath, 0, (CHAR *) 0);
	return(0);
}


#define	NBUCKETS		20	/* number of buckets in histogram */

typedef struct
{
	TimeType  begin, end, size;
	long  table[NBUCKETS];
} Accounts;


/*
 * helper routine for activity command
 */
static int adoit(nptr n, Accounts *ac)
{
	if (! (n->nflags & (ALIAS | MERGED | POWER_RAIL)))
	{
		if ((long)n->c.time >= ac->begin && (long)n->c.time <= ac->end)
			ac->table[(n->c.time - ac->begin) / ac->size] += 1;
	}
	return(0);
}


/*
 * print histogram of circuit activity in specified time interval
 */
static int doactivity(void)
{
	register int  i;
	static CHAR   st[] = {x_("**************************************************")};
	Accounts      ac;
	long          total;

#   define	SIZE_ST		(sizeof(st) - 1)

	if (targc == 2)
	{
		ac.begin = (TimeType)ns2d(eatof(targv[1]));
		ac.end = irsim_cur_delta;
	} else
	{
		ac.begin = (TimeType)ns2d(eatof(targv[1]));
		ac.end = (TimeType)ns2d(eatof(targv[2]));
	}

	if (ac.end < ac.begin)
		SWAP(TimeType, ac.end, ac.begin);

	/* collect histogram info by walking the network */
	for(i = 0; i < NBUCKETS; ac.table[i++] = 0);

	ac.size = (ac.end - ac.begin + 1) / NBUCKETS;
	if (ac.size <= 0)
		ac.size = 1;

	irsim_walk_net1((int(*)(nptr, CHAR*))adoit, (CHAR *)&ac);

	/* print out what we found */
	for(total = i = 0; i < NBUCKETS; i++) total += ac.table[i];

	ttyputmsg(x_("Histogram of circuit activity: %.2f -> %.2fns (bucket size = %.2f)"),
		d2ns(ac.begin), d2ns(ac.end), d2ns(ac.size));

	for(i = 0; i < NBUCKETS; i += 1)
		ttyputmsg(x_(" %10.2f -%10.2f%6d  %s"),
			d2ns(ac.begin + (i * ac.size)), d2ns(ac.begin + (i + 1) * ac.size),
				ac.table[i], &st[SIZE_ST - (SIZE_ST * ac.table[i]) / total]);

	return(0);
}


/*
 * Helper routine for "changes" command.
 */
static int cdoit(nptr n, Accounts *ac)
{
	int   i;

	UnAlias(n);

	if (n->nflags & (MERGED | ALIAS))
		return(0);

	if ((long)n->c.time >= ac->begin && (long)n->c.time <= ac->end)
	{
		i = estrlen(pnode(n)) + 2;
		if (column + i >= MAXCOL)
		{
			column = 0;
		}
		column += i;
		ttyputmsg(x_("  %s"), pnode(n));
	}
	return(0);
}


/*
 * Print list of nodes which last changed value in specified time interval
 */
static int dochanges(void)
{
	Accounts  ac;

	if (targc == 2)
	{
		ac.begin = (TimeType)ns2d(eatof(targv[1]));
		ac.end = irsim_cur_delta;
	} else
	{
		ac.begin = (TimeType)ns2d(eatof(targv[1]));
		ac.end = (TimeType)ns2d(eatof(targv[2]));
	}

	column = 0;
	ttyputmsg(x_("Nodes with last transition in interval %.2f -> %.2fns:"),
		d2ns(ac.begin), d2ns(ac.end));

	irsim_walk_net1((int(*)(nptr, CHAR*))cdoit, (CHAR *)&ac);
	if (column != 0)
		return(0);
	return(0);
}


/*
 * Helper routine for "printx" command.
 */
static int xdoit(nptr n)
{
	int   i;

	UnAlias(n);

	if (! (n->nflags & (MERGED | ALIAS)) && n->npot == X)
	{
		i = estrlen(pnode(n)) + 2;
		if (column + i >= MAXCOL)
		{
			column = 0;
		}
		column += i;
		ttyputmsg(x_("  %s"), pnode(n));
	}
	return(0);
}


/*
 * Print list of nodes with undefined (X) value
 */
static int doprintX(void)
{
	ttyputmsg(x_("Nodes with undefined potential:"));
	column = 0;
	irsim_walk_net(xdoit);
	if (column != 0)
		return(0);
	return(0);
}


/*
 * Helper routine for printing aliases
 */
static int aldoit(nptr n)
{
	CHAR  *nname = pnode(n);
	CHAR  *is_merge;

	if (n->nflags & ALIAS)
	{
		UnAlias(n);
		is_merge = (CHAR *)((n->nflags & MERGED) ? x_(" (part of a stack)") : x_(""));
		ttyputmsg(x_("  %s -> %s%s"), nname, pnode(n), is_merge);
	}
	return(0);
}


/*
 * Print nodes that are aliases
 */
static int doprintAlias(void)
{
	if (irsim_naliases == 0)
		ttyputmsg(x_("there are no aliases"));
	else
	{
		ttyputmsg(x_("there are %d aliases:"), irsim_naliases);
		irsim_walk_net(aldoit);
	}
	return(0);
}


/*
 * Helper routine to print pending events
 */
static int print_list(int n, evptr l, evptr eolist)
{
	if (l == NULL)
		return(n);
	for(eolist = eolist->flink; l != eolist && n != 0; l = l->flink, n--)
	{
		ttyputmsg(x_("Node %s -> %c @ %.2fns (%.2fns)"),
			pnode(l->enode), irsim_vchars[l->eval], d2ns(l->ntime),
				d2ns(l->ntime - irsim_cur_delta));
	}
	return(n);
}


/*
 * Print list of pending events
 */
static int printPending(void)
{
	int    n;
	TimeType   delta = 0;
	evptr  list, eolst;

	n = (targc == 2) ? eatoi(targv[1]) : -1;

	while ((delta = irsim_pending_events(delta, &list, &eolst)) && n != 0)
		n = print_list(n, list, eolst);
	n = print_list(n, list, eolst);
	return(0);
}


/*
 * set/reset various display parameters
 */
static int dodisplay(void)
{
	static CHAR    cmdfile_str[] = {x_("cmdfile")}, automatic_str[] = {x_("automatic")};
	register int   i, value;
	register CHAR  *p;

	if (targc == 1)
	{
		ttyputmsg(x_("display = %s%s %s%s"),
			dcmdfile ? x_("") : x_("-"), cmdfile_str,
				ddisplay ? x_("") : x_("-"), automatic_str);
		return(0);
	}

	for(i = 1; i < targc; i += 1)
	{
		p = targv[i];
		if (*p == '-')
		{
			value = 0;
			p += 1;
		}
		else
			value = 1;

		if (namesame(p, cmdfile_str) == 0)
			dcmdfile = value;
		else if (namesame(p, automatic_str) == 0)
			ddisplay = value;
		else
			irsim_error(filename, lineno, x_("unrecognized display parameter: %s"),
				targv[i]);
	}
	return(0);
}


/*
 * Print list of transistors with src/drn shorted (or between power supplies).
 */
static int print_tcap(void)
{
	tptr  t;

	if (irsim_tcap->scache.t == irsim_tcap)
		ttyputmsg(x_("there are no shorted transistors"));
	else
		ttyputmsg(x_("shorted transistors:"));
	for(t = irsim_tcap->scache.t; t != irsim_tcap; t = t->scache.t)
	{
		ttyputmsg(x_(" %s g=%s s=%s d=%s (%gx%g)"),
			irsim_ttype[BASETYPE(t->ttype)], 
				pnode(t->gate), pnode(t->source), pnode(t->drain),
					t->r->length / (double) irsim_LAMBDACM, t->r->width / (double) irsim_LAMBDACM);
	}
	return(0);
}


static int set_incres(void)
{
	if (targc == 1)
		ttyputmsg(x_("incremental resolution = %.2f"), d2ns(irsim_INC_RES));
	else
	{
		long  new_res = (long) ns2d(eatof(targv[1]));

		if (new_res < 0)
			irsim_error(filename, lineno, x_("resolution must be positive"));
		else
			irsim_INC_RES = new_res;
	}
	return(0);
}


static	CHAR	*changelog = NULL;


/*
 * set the changes-log filename. 
 */
static int setlogchanges(void)
{
	INTBIG filestatus;

	if (targc == 1)
	{
		ttyputmsg(x_("changes-logfile is %s"),
			(changelog == NULL) ? x_("turned OFF") : changelog);
	}
	else
	{
		if (namesame(x_("off"), targv[1]) == 0)
		{
			if (changelog != NULL)
			{
				efree(changelog);
				changelog = NULL;
			}
		}
		else
		{
			filestatus = fileexistence(targv[1]);
			if (filestatus == 2 || filestatus == 3)
				ttyputmsg(x_("can't write to '%s'"), targv[1]);
			else
			{
				if (filestatus == 0)
					ttyputmsg(x_("OK, starting a new log file"));
				else
					ttyputmsg(x_("%s already exists, will append to it"),
						targv[1]);
				if (changelog != NULL)
					efree(changelog);
				changelog = (CHAR *)emalloc((estrlen(targv[1]) + 1) * SIZEOFCHAR, sim_tool->cluster);
				if (changelog != NULL)
					(void)estrcpy(changelog, targv[1]);
				else
					ttyputmsg(x_("out of memory, logfile is OFF"));
			}
		}
	}
	return(0);
}


/*
 * read changes and do incremental simulation.
 */
static int do_incsim(void)
{
	iptr  ch_list = NULL;

	CHECK_STOP();

	if (irsim_sim_time0 != 0)
	{
		ttyputmsg(x_("Warning: part of the history was flushed:"));
		ttyputmsg(x_("         incremental results may be wrong"));
	}

	ch_list = irsim_rd_changes(targv[1], changelog);

	if (ch_list == NULL)
		ttyputmsg(x_("no affected nodes: done"));
	else
		irsim_incsim(ch_list);
	if (dodisplay())
		pnwatchlist();
	else
		prtime(0);
	return(0);
}


/*
 * Startup analyzer window and/or add more signals to analyzer display list.
 */
static int analyzer(void)
{
	if (! irsim_analyzerON)
	{
		if (! irsim_InitDisplay()) return(0);
		irsim_InitTimes(irsim_sim_time0, stepsize, irsim_cur_delta);
	}

	if (targc > 1)
	{
		int  ndigits = 0;

		if (targv[1][0] == '-' && targv[1][2] == '\0')
		{
			switch(targv[1][1])
			{
				case 'b':	ndigits = 1;	shift_args(TRUE);	break;
				case 'o':	ndigits = 3;	shift_args(TRUE);	break;
				case 'h':	ndigits = 4;	shift_args(TRUE);	break;
				default: ;
			}
		}

		if (targc > 1)
			apply((int(*)())irsim_AddNode, (int(*)())irsim_AddVector, (CHAR *) &ndigits);
	}

	irsim_UpdateWindow(0);
	irsim_DisplayTraces(irsim_analyzerON);		/* pass 0 first time */

	irsim_analyzerON = TRUE;
	return(0);
}


/*
 * Clear the analyzer display list.
 */
static int clear_analyzer(void)
{
	if (irsim_analyzerON)
		irsim_ClearTraces();
	else
		ttyputmsg(x_("analyzer is off"));
	return(0);
}


/*
 * Write out entire history to a file.
 */
static int dump_hist(void)
{
	CHAR  fname[256];

	if (first_file == NULL || irsim_cur_delta == 0)
	{
		irsim_error(filename, lineno, x_("Nothing to dump"));
		return(0);
	}

	if (targc == 1)
		(void)esnprintf(fname, 256, x_("%s.hist"), first_file);
	else
		(void)estrcpy(fname, targv[1]);

	irsim_DumpHist(fname);

	return(0);
}


/*
 * Read network history from a file.
 */
static int do_readh(void)
{
	if (irsim_analyzerON)
		irsim_StopAnalyzer();
	irsim_ReadHist(targv[1]);
	if (irsim_analyzerON)
		irsim_RestartAnalyzer(irsim_sim_time0, irsim_cur_delta, TRUE);
	return(0);
}


/*
 * Move back simulation time to specified time.
 */
static int back_time(void)
{
	TimeType  newt;

	CHECK_STOP();

	newt = (TimeType)ns2d(eatof(targv[1]));
	if (newt < irsim_sim_time0 || newt > irsim_cur_delta)
	{
		irsim_error(filename, lineno, x_("%s: invalid time"), targv[1]);
		return(0);
	}
	if (irsim_analyzerON)
		irsim_StopAnalyzer();

	irsim_cur_delta = newt;
	irsim_ClearInputs();
	(void)irsim_back_sim_time(irsim_cur_delta, FALSE);
	irsim_cur_node = NULL;			/* fudge */
	irsim_walk_net(irsim_backToTime);
	if (irsim_cur_delta == 0)
		irsim_ReInit();

	if (irsim_analyzerON)
		irsim_RestartAnalyzer(irsim_sim_time0, irsim_cur_delta, TRUE);

	pnwatchlist();
	return(0);
}


/*
 * Write network to file.
 */
static int wr_net(void)
{
	CHAR  fname[256];

	if (first_file == NULL)
	{
		irsim_error(filename, lineno, x_("No network?"));
		return(0);
	}

	if (targc == 1)
		(void)esnprintf(fname, 256, x_("%s.inet"), first_file);
	else
		(void)estrcpy(fname, targv[1]);

	irsim_wr_netfile(fname);

	return(0);
}


#ifdef STATS

int  ev_hgm = 0;
static	struct
{
	hptr  head;
	hptr  tail;
} ev_hgm_table[5];

void IncHistEvCnt(int tp)
{
	hptr  h;
	long  tm;
	int   indx;

	if (ev_hgm == 0) return;

	switch(tp)
	{
		case -1:        indx = 0;	break;
		case PUNTED:
		case REVAL:
		case DECAY_EV:  indx = 1;	break;
		case STIM_INP:
		case STIM_XINP:
		case STIMULI:   indx = 2;	break;
		case CHECK_PNT: indx = 3;	break;
		case XINP_EV:
		case INP_EV:    indx = 4;	break;
		default:					return;
	}

	tm = irsim_cur_delta / 10;
	h = ev_hgm_table[indx].tail;
	if (h->time == tm)
		h->t.xx += 1;
	else
	{
		if ((h = irsim_freeHist) == NULL)
			h = (hptr) irsim_MallocList(sizeof(HistEnt), 1);
		irsim_freeHist = h->next;
		if (ev_hgm_table[indx].tail == irsim_last_hist)
			ev_hgm_table[indx].head = h;
		else
			ev_hgm_table[indx].tail->next = h;
		ev_hgm_table[indx].tail = h;
		h->next = irsim_last_hist;
		h->time = tm;
		h->t.xx = 1;
	}
}


static int do_ev_stats(void)
{
	int   i;

	if (targc == 1)
	{
		ttyputmsg(x_("event recording is %s"), (ev_hgm) ? x_("ON") : x_("OFF"));
		return(0);
	}

	if (namesame(x_("on"), targv[1]) == 0)
	{
		static int last = 5;

		ev_hgm = 1;
		for(i = 0; i < last; i++)
			ev_hgm_table[i].head = ev_hgm_table[i].tail = irsim_last_hist;
		last = 0;
	}
	else if (namesame(x_("clear"), targv[1]) == 0)
	{
		for(i = 0; i < 5; i++)
			ev_hgm_table[i].head = ev_hgm_table[i].tail = irsim_last_hist;
	}
	else if (namesame(x_("off"), targv[1]) == 0)
		ev_hgm = 0;
	else
		irsim_error(filename, lineno, x_("don't know what '%s' means"), targv[1]);

	return(0);
}


static do_pr_ev_stats(void)
{
	static CHAR *ev_name[] =
	{ x_("evaluation"), x_("I-evaluation"), x_("stimulus"), x_("check-point"), x_("input") };
	FILE       *fp;
	int        i, lim, j;
	hptr       h;

	if (targc == 2)
	{
		fp = efopen(targv[1], x_("w"));
		if (fp == NULL)
		{
			irsim_error(filename, lineno, x_("cannot open file '%s'"), targv[1]);
			return(0);
		}
	}
	else
		fp = stdout;

	efprintf(fp, x_("Event Activity"));

	lim = (irsim_i_nevals != 0) ? 5 : 1;

	for(i = j = 0; i < lim; i++)
	{
		h = ev_hgm_table[i].head;
		if (h == irsim_last_hist) continue;
		j++;
		efprintf(fp, x_("\n** %s:\n"), ev_name[i]);
		while (h != irsim_last_hist)
		{
			efprintf(fp, x_("%d\t%d\n"), h->time, h->t.xx);
			h = h->next;
		}
		efprintf(fp, x_("\n"));
	}
	if (j == 0)
	{
		efprintf(fp, x_(": Nothing Recorded\n"));
		if (targc == 2) ttyputmsg(x_(": Nothing Recorded"));
	}
	if (targc == 2)
		fclose(fp);
	return(0);
}

#endif

#ifdef CL_STATS

static	int CLcount[1001];

RecordConnList(int ntx)
{
	if (ntx > 1000) ntx = 1000;
	CLcount[ntx] += 1;
}

static int cl_compar(short *a, short *b)
{
	return(CLcount[*b] - CLcount[*a]);
}

int do_cl_stats(void)
{
	short   cnt_indx[1001];
	FILE    *fp;
	int     i, ch, tot, n;
	double  avg, dev;

	if (targc == 2)
	{
		fp = efopen(targv[1], x_("w"));
		if (fp == NULL)
		{
			irsim_error(filename, lineno, x_("cannot open file '%s'"), targv[1]);
			return(0);
		}
	}
	else
		fp = stdout;

	for(avg = 0, tot = i = 0; i <= 1000; i++)
	{
		cnt_indx[i] = i;
		if (CLcount[i] > 0)
		{
			avg += i * CLcount[i];
			tot += CLcount[i];
		}
	}
	avg = avg / tot;

	for(dev = 0, i = 0; i <= 1000; i++)
	{
		if (CLcount[i] > 0)
			dev += CLcount[i] * (i - avg) * (i - avg);
	}
	dev = sqrt(dev / tot);

	qsort(cnt_indx, 1001, sizeof(short), cl_compar);

	efprintf(fp, x_("Connection-list statistics\n"));
	efprintf(fp, x_("\tavg-num-trans = %.2f  std-deviation = %.2f\n"), avg, dev);

	efprintf(fp, x_("num-trans  num-times      %%  %%accum\n"));
	efprintf(fp, x_("---------  ---------  -----  ------\n"));
	avg = tot;
	dev = 0;
	for(i = 0; i <= 1000; i++)
	{
		double  p;

		n = cnt_indx[i];
		if ((tot = CLcount[n]) == 0)
			continue;
		ch = (n == 1000) ? '>' : ' ';
		p = 100.0 * tot / avg;
		dev += p;
		efprintf(fp, x_("%c%8d  %9d  %5.2f  %6.2f\n"), ch, n, tot, p, dev);
	}

	if (targc == 2)
		fclose(fp);
	return(0);
}

#endif


typedef struct
{
	int  nsd, ng;
} TranCnt;

static int count_trans(nptr n, TranCnt *tt)
{
	register lptr  l;
	register int   i;

	if ((n->nflags & (ALIAS | POWER_RAIL)) == 0)
	{
		for(i = 0, l = n->ngate; l != NULL; l = l->next, i++);
			tt->ng += i;
		for(i = 0, l = n->nterm; l != NULL; l = l->next, i++);
			tt->nsd += i;
	}
	return(0);
}


/*
 * Print event statistics.
 */
static int do_stats(void)
{
	CHAR  n1[10], n2[10];

	if (targc == 2)
	{
		static TranCnt  tt = { 0, 0 };

		if (tt.ng == 0 && tt.nsd == 0)
		{
			irsim_walk_net1((int(*)(nptr, CHAR*))count_trans, (CHAR *)&tt);
			ttyputmsg(x_("avg: # gates/node = %g,  # src-drn/node = %g"),
				(double) tt.ng / irsim_nnodes, (double) tt.nsd / irsim_nnodes);
		}
	}
	ttyputmsg(x_("changes = %d"), irsim_num_edges);
	ttyputmsg(x_("punts (cns) = %d (%d)"),
	irsim_num_punted, irsim_num_cons_punted);
	if (irsim_num_punted == 0)
	{
		(void)estrcpy(n1, x_("0.0"));
		(void)estrcpy(n2, n1);
	} else
	{
		(void)esnprintf(n1, 10, x_("%2.2f"),
			100.0 / ((float) irsim_num_edges / irsim_num_punted + 1.0));
		(void)esnprintf(n2, 10, x_("%2.2f"),
			(float) (irsim_num_cons_punted * 100.0 /irsim_num_punted));
	}
	ttyputmsg(x_("punts = %s%%, cons_punted = %s%%"), n1, n2);

	ttyputmsg(x_("nevents = %ld; evaluations = %ld"), irsim_nevent, irsim_nevals);
	if (irsim_i_nevals != 0)
	{
		ttyputmsg(x_("inc. evaluations = %ld; events:"), irsim_i_nevals);
		ttyputmsg(x_("reval:      %ld"), irsim_nreval_ev);
		ttyputmsg(x_("punted:     %ld"), irsim_npunted_ev);
		ttyputmsg(x_("stimuli:    %ld"), irsim_nstimuli_ev);
		ttyputmsg(x_("check pnt:  %ld"), irsim_ncheckpt_ev);
		ttyputmsg(x_("delay chk:  %ld"), irsim_ndelaychk_ev);
		ttyputmsg(x_("delay ev:   %ld"), irsim_ndelay_ev);
	}

	return(0);
}


/*
 * Shift the command left/right by 1.
 */
static void shift_args(int left)
{
	register int   ac;
	register CHAR  **ap;
	register CHAR  *wp;

	if (left)
	{
		targc--;
		for(ac = 0, ap = targv, wp = wildCard; ac < targc; ac++)
			ap[ac] = ap[ac + 1],    wp[ac] = wp[ac + 1];
	} else
	{
		for(ac = targc++, ap = targv, wp = wildCard; ac >= 0; ac--)
			ap[ac + 1] = ap[ac],    wp[ac + 1] = wp[ac];
	}
}


/*
 * time a command and print resource utilization.
 */
static int do_time(void)
{
	CHAR  usage_str[40];
	int   narg, i = 0;

	ttyputerr(x_("unable to time function..."));
	shift_args(TRUE);
	narg = targc;
	if (narg)
		i = exec_cmd();
	ttyputmsg(x_("%s"), usage_str);
	return(i);
}

static int doHist(void)
{
	if (targc == 1)
	{
		ttyputmsg(x_("History is "));
		if (irsim_sm_stat && OUT_OF_MEM) ttyputmsg(x_("off."));
		else ttyputmsg(x_("on."));
	} else
	{
		if (estrcmp(targv[1], x_("on")) != 0)
		{
			irsim_sm_stat |= OUT_OF_MEM; }
			else { irsim_sm_stat &= ~OUT_OF_MEM;
		}
	}
	return (0);
}

/*
 * Flush out the recorded history up to the (optional) specified time.
 */
static int flush_hist(void)
{
	TimeType  ftime;

	if (targc == 1)
		ftime = irsim_cur_delta;
	else
	{
		ftime = ns2d(eatof(targv[1]));
		if (ftime < 0 || ftime > irsim_cur_delta)
		{
			irsim_error(filename, lineno, x_("%s: Invalid flush time"), targv[1]);
			return(0);
		}
	}

	if (ftime == 0)
		return(0);

	if (irsim_analyzerON)
		irsim_StopAnalyzer();

	irsim_FlushHist(ftime);
	irsim_sim_time0 = ftime;

	if (irsim_analyzerON)
		irsim_RestartAnalyzer(irsim_sim_time0, irsim_cur_delta, TRUE);

	return(0);
}


/*
 * Just print the string YES if transistor coordinates are avalible.
 * This is used by rsim magic interface to use transistor positions to
 * identify nodes.
 */
static int HasCoords(void)
{
	if (irsim_txt_coords)
		ttyputmsg(x_("YES"));
	return(0);
}


#ifdef FAULT_SIM

static int parse_trigger(void)
{
	int       edge;
	TimeType      delay;
	Find1Arg  f;

	if ((targc < 3 || targc > 4))
		goto bad_trigger;

	delay = (targc > 3) ? ns2d(eatof(targv[3])) : 0;

	edge = ch2pot(targv[2][0]);
	if (edge >= N_POTS)
		return(1);
	if (edge != LOW && edge != HIGH)
		goto bad_trigger;

	FindOne(&f);
	if (f.num > 1 || f.vec != NULL)
	{
		irsim_error(filename, lineno, x_("%s: not a single node"), targv[1]);
		return(1);
	}
	if (irsim_add_trigger(f.node, edge, delay))
	{
		irsim_error(filename, lineno, x_("trigger: %s has no %s transitions"),
		pnode(f.node), (edge == LOW) ? x_("1 -> 0") : x_("1 -> 0 "));
	}
	return(0);

bad_trigger:
	irsim_error(filename, lineno, x_("expected: \"trigger\" node 0|1 [delay]"));
	return(1);
}


static int parse_sampler(void)
{
	TimeType  period, offset = 0;

	if (targc < 2 || targc > 3)
	goto bad_sample;

	period = ns2d(eatof(targv[1]));
	if (period <= 0)
	{
		irsim_error(filename, lineno, x_("%s: Illegal period"), targv[1]);
		return(1);
	}
	if (targc == 3)
	{
		offset = ns2d(eatof(targv[2]));
		if (offset < 0)
		goto bad_sample;
	}

	if (irsim_cur_delta <= offset)
	{
		irsim_error(filename, lineno, x_("can't sample, simulation time too small"));
		return(1);
	}

	return(irsim_add_sampler(period, offset));

bad_sample:
	irsim_error(filename, lineno, x_("expected: \"sample\" period [offset]"));
	return(1);
}


static int setup_fsim(CHAR *file, int *p_seed)
{
	FILE      *fp;
	CHAR      line[256], *dummy;
	int       olineno = lineno;
	CHAR      *ofname = filename;
	int       doing_outputs, err, percent, look_p;

	fp = xopen(file, el_filetypetext, 0, &dummy);
	if (fp == NULL)
	{
		irsim_error(filename, lineno, x_("cannot open '%s'"), file);
		return(1);
	}

	filename = file;
	lineno = 0;
	err = percent = 0;
	doing_outputs = FALSE;
	look_p = TRUE;

	while (! err && !xfgets(line, 256, fp))
	{
		lineno += 1;
		parse_line(line, 256);
		if (targc == 0)
			continue;

		if (look_p)	/* seed must be 1st non-empty line in file */
		{
			look_p = FALSE;
			if (namesame(x_("seed"), targv[0]) == 0)
			{
				if (targc < 2)
				{
					irsim_error(file, lineno, x_("syntax: \"seed\" <percentage>"));
					err = 1;
				} else
				{
					percent = eatoi(targv[1]);
					if (percent <= 0 || percent > 100)
					{
						irsim_error(file, lineno,
							x_("percentage must be in the range [1-100]"));
						err = 1;
					}
				}
				continue;
			}
		}

		if (! doing_outputs)
		{
			if (namesame(x_("sample"), targv[0]) == 0)
				err = parse_sampler();
			else if (namesame(x_("trigger"), targv[0]) == 0)
				err = parse_trigger();
			else
			{
				irsim_error(file, lineno, x_("expected: \"trigger\" or \"sample\""));
				err = 1;
			}
			doing_outputs = TRUE;
		}
		else if (targc == 1 && estrcmp(x_("***"), targv[0]) == 0)
		{
			doing_outputs = FALSE;
		} else
		{
			int  any = 0;
			shift_args(FALSE);
			apply((int(*)())irsim_add_prim_output, 0, (CHAR *) &any);
			if (any != 1)
				err = 1;
		}
	}

	(void)fclose(fp);
	filename = ofname;
	lineno = olineno;
	*p_seed = percent;
	return(err);
}


static int do_fsim(void)
{
	int   p_seed;
	CHAR  *outname;

	CHECK_STOP();

	if (irsim_cur_delta == 0)
	{
		ttyputmsg(x_("Circuit needs to be simulated before faultsim"));
		return(0);
	}
	if (irsim_sim_time0 != 0)
	{
		ttyputmsg(x_("Can't faultsim: Incomplete history"));
		return(0);
	}

	outname = (targc == 3) ? targv[2] : NULL;

	if (setup_fsim(targv[1], &p_seed) == 0)
		irsim_exec_fsim(outname, p_seed);

	irsim_cleanup_fsim();

	return(0);
}
#endif


static	int do_help(void);	/* forward reference */


/* list of commands and their handlers */
static Command  cmds[] = 
{
	{ x_("!"),		quest,		2,	MAXARGS,
		x_("node/vector... -> info regarding node(s) gate connections")	},
	{ x_("<"),		do_rdstate,	2,	2,
		x_("file -> restore network state from file")				},
	{ x_("<<"),		do_rdstate,	2,	2,
		x_("file -> restore network and inputs state from file")		},
	{ x_(">"),		do_wrstate,	2,	2,
		x_("file -> write current network state to file")			},
	{ x_("?"),		quest,		2,	MAXARGS,
		x_("node/vector... -> info regarding node(s) src/drn connections")	},
	{ x_("@"),		cmdfile,	2,	2,
		x_("file -> read commands from command file")				},
	{ x_("activity"),	doactivity,	2,	3,
		x_("from [to] -> circuit activity in time interval")			},
	{ x_("alias"),		doprintAlias,	1,	1,
		x_(" ->  print node aliases")						},
	{ x_("ana"),		analyzer,	1,	MAXARGS,
		x_("shorthand for \"analyzer\"")					},
	{ x_("analyzer"),	analyzer,	1,	MAXARGS,
		x_("[-b|-o|-h] node/vector... -> display node/vector(s) in analyzer")	},
	{ x_("assert"),		doAssert,	3,	4,
		x_("node/vector [mask] val -> assert node/vector = val [& mask = 0]")	},
	{ x_("assertWhen"),		doAssertWhen,	5,	5,
		x_("nodeT valT node val -> assert node = val when nodeT switches to valT")},
	{ x_("back"),		back_time,	2,	2,
		x_("time -> move simulation time back to specified to time")		},
	{ x_("c"),		doclock,	1,	2,
		x_("[n] -> simulate for n clock cycles (default 1)\n			-> or continue last simulation command prior to stoping")		},
	{ x_("changes"),	dochanges,	2,	3,
		x_("from [to] -> print nodes that changed in time interval")		},
	{ x_("clear"),		clear_analyzer,	1,	1,
		x_(" -> remove all signals from analyzer window")			},
	{ x_("clock"),		setclock,	1,	MAXARGS,
		x_("[node/vector [val]] -> define clock sequence for node/vector")	},
	{ x_("d"),		pnlist,		1,	MAXARGS,
		x_("[node/vector]... -> print node/vector(s) or display-list")	},
	{ x_("debug"),		setdbg,		1,	MAXARGS,
		x_("[options] -> print/set debug state (? for help)")			},
	{ x_("decay"),		setdecay,	1,	2,
		x_("[file] -> write net history to file")				},
	{ x_("display"),	dodisplay,	1,	3,
		x_("[[-]option] -> print/set cmdfile and auto-watch-list display")	},
	{ x_("dumph"),		dump_hist,	1,	2,
		x_("[file] -> write network state history to file")			},
	{ x_("exit"),		doexit,		1,	2,
		x_("[status] -> exit program with given status (default: 0)")		},
	{ x_("flush"),		flush_hist,	1,	2,
		x_("[time] -> flush out history up to time (default:current-time)")	},
	{ x_("h"),		setvalue,	2,	MAXARGS,
		x_("node/vector... -> drive node/vector(s) to 1 (high)")		},
	{ x_("has_coords"),	HasCoords,	1,	1,
		x_(" -> print if transistor coordinates are available")		},
#ifdef STATS
	{ x_("histev"),		do_ev_stats,	1,	2,
		x_("[clear | off | on] -> enable/disable/clear event activity record")},
	{ x_("evstats"),	do_pr_ev_stats,	1,	2,
		x_("[file] -> print event activity recorded")				},
#endif
#ifdef CL_STATS
	{ x_("clstats"),	do_cl_stats,	1,	2,
		x_("[file] -> print connection-list statistics")			},
#endif
	{ x_("help"),		do_help,	1,	MAXARGS,
		x_("[command]... -> print info on command(s) or available commands")	},
	{ x_("hist"),           doHist,         1,      2,
		x_("[on/off] -> display or set history collection mode")             },
	{ x_("inputs"),		inputs,		1,	1,
		x_(" -> print currently driven (input) nodes")			},
	{ x_("ires"),		set_incres,	1,	2,
		x_("[time] -> print/set incremental resolution to time")		},
	{ x_("isim"),		do_incsim,	2,	2,
		x_("[file] -> read changes from file and incrementally resimulate")	},
	{ x_("l"),		setvalue,	2,	MAXARGS,
		x_("node/vector... -> drive node/vector(s) to 0 (low)")		},
	{ x_("logfile"),	setlog,		1,	2,
		x_("[[+]file] -> start/stop log file (+file appends to file)")	},
	{ x_("model"),		setmodel,	1,	2,
		x_("[linear|switch] -> print/change simulation model")		},
	{ x_("p"),		dophase,	1,	1,
		x_("step clock one simulation step (phase)")				},
	{ x_("path"),		dopath,		2,	MAXARGS,
		x_("node/vector... -> critical path for last transition of node(s)")	},
	{ x_("print"),		domsg,		1,	MAXARGS,
		x_("[text...] -> print specified text")				},
	{ x_("printp"),		printPending,	1,	2,
		x_("[n] -> print up to 'n' pending events (default: all)")		},
	{ x_("printx"),		doprintX,	1,	1,
		x_(" -> print all undefined (X) nodes")				},
	{ x_("q"),		quit,		1,	1,
		x_(" -> terminate input from current stream")				},
	{ x_("quit"),		doexit,		1,	2,
		x_("synonym for exit")		},
	{ x_("R"),		runseq,		1,	2,
		x_("[n] -> simulate for 'n' cycles (default: longest sequence)")	},
	{ x_("readh"),		do_readh,	2,	2,
		x_("file -> read network state history from file")			},
	{ x_("report"),		setreport,	1,	10,
		x_("[args] -> print/set trace-info or decay report (? for help)")	},
	{ x_("s"),		dostep,		1,	2,
		x_("[time] -> simulate for specified time (default: stepsize)")	},
	{ x_("set"),		setvector,	3,	3,
		x_("vector value -> assign value to vector")				},
	{ x_("setlog"),		setlogchanges,	1,	2,
		x_("[file|off] -> print/change net-changes log-filename")		},
	{ x_("setpath"),	docmdpath,	1,	MAXARGS,
		x_("[[+] path]... -> print/set/add-to cmd files search path")		},
	{ x_("stats"),		do_stats,	1,	2,
		x_(" -> print event statistics")					},
	{ x_("stepsize"),	setstep,	1,	2,
		x_("[time] -> print/set simulation step size")			},
	{ x_("stop"),		setstop,	2,	MAXARGS,
		x_("[-]node/vector... -> pause simulation when node/vector(s) change")},
	{ x_("t"),		settrace,	2,	MAXARGS,
		x_("[-]node/vector... -> start/stop tracing specified node/vector(s)")},
	{ x_("tcap"),		print_tcap,	1,	1,
		x_(" -> print all shorted transistors")				},
	{ x_("time"),		do_time,	1,	MAXARGS,
		x_("[command...] -> time given command or program")			},
	{ x_("u"),		setvalue,	2,	MAXARGS,
		x_("[node/vector]... -> drive a node/vector(s) to X (undefined)")	},
	{ x_("unitdelay"),	setunit,	1,	2,
		x_("[time] -> force transitions to specified time (0 to disable)")	},
	{ x_("until"),          doUntil,        4,      5,
		x_("node/vec [mask] val count -> sim until = val [& mask = 0] or count runout")	},
	{ x_("V"),		setseq,		1,	MAXARGS,
		x_("[node/vector [val...]] -> define input sequence for node/vector")	},
	{ x_("vector"),		dovector,	3,	MAXARGS,
		x_("name node... -> (re)define vector 'name' composed of node(s)")	},
	{ x_("w"),		display,	2,	MAXARGS,
		x_("[-]node/vector... -> add/delete node/vector(s) to display-list")	},
	{ x_("wnet"),		wr_net,		1,	MAXARGS,
		x_("[file] -> write network to file (smaller/faster than sim file)")	},
	{ x_("x"),		setvalue,	2,	MAXARGS,
		x_("node/vector... -> make node/vector(s) undriven (non-input)")	},
#ifdef FAULT_SIM
	{ x_("faultsim"),	do_fsim,	2,	3,
		x_("infile [outfile] -> do stuck-at fault simulation")		},
#endif
#ifdef POWER_EST
	{ x_("powlogfile"),	setcaplog,	1,	2,
		x_("[[+]file] -> start/stop power logfile (+file appends to file)")	},
	{ x_("powtrace"),	setpowtrace,	2,	MAXARGS,
		x_("[-]node/vector... -> start/stop power tracing specified node/vector(s)")},
	{ x_("sumcap"),		sumcap,		1,	2,
		x_(" -> print out sum of capacitances of all nodes")			},
	{ x_("vsupply"),	setvsupply,	1,	2,
		x_("[v] Set supply voltage = v Volts (no arguments displays value)")	},
	{ x_("powstep"),	togglepstep,	1,	1,			
		x_(" -> Toggle display of power estimate for each step")		},
#endif
	{ NULL,		NULL,		0,	0,	NULL		}
};


static int do_help(void)
{
	Command  *c;
	int      i, col;
	CHAR line[MAXCOL];

	if (targc == 1)
	{
		estrcpy(line, x_("available commands:"));
		col = estrlen(line);
		for(c = cmds; c->name != NULL; c++)
		{
			i = estrlen(c->name) + 1;
			if (col + i >= MAXCOL)
			{
				ttyputmsg(x_("%s"), line);
				line[0] = 0;
				col = 0;
			}
			col += i;
			estrcat(line, x_(" "));
			estrcat(line, c->name);
		}
		if (col > 0) ttyputmsg(x_("%s"), line);
	} else
	{
		for(i = 1; i < targc; i++)
		{
			for(c = cmdtbl[HashCmd(targv[i])]; c != NULL; c = c->next)
			{
				if (estrcmp(targv[i], c->name) == 0)
				break;
			}
			if (c)
				ttyputmsg(x_("%s %s"), c->name, c->help);
			else
				ttyputmsg(x_("%s -> UNKNOWN"), targv[i]);
		}
	}
	return(0);
}


static int cmd_cmp(const void *ap, const void *bp)
{
	register Command *a, *b;

	a = (Command *)ap;
	b = (Command *)bp;
	return(namesame(a->name, b->name));
}


void irsim_init_commands(void)
{
	register int      n;
	register Command  *c;

	n = sizeof(cmds) / sizeof(Command) - 1;
	qsort(cmds, n, sizeof(Command), cmd_cmp);

	for(n = 0; n < CMDTBLSIZE; n++)
		cmdtbl[n] = NULL;
	for(c = cmds; c->name != NULL; c += 1)    
	{
		n = HashCmd(c->name);
		c->next = cmdtbl[n];
		cmdtbl[n] = c;
	}
}
