/* -*- tab-width: 4 -*-
 *
 *     ********************************************************************* 
 *     * Copyright (C) 1988, 1990 Stanford University.                     * 
 *     * Permission to use, copy, modify, and distribute this              * 
 *     * software and its documentation for any purpose and without        * 
 *     * fee is hereby granted, provided that the above copyright          * 
 *     * notice appear in all copies.  Stanford University                 * 
 *     * makes no representations about the suitability of this            * 
 *     * software for any purpose.  It is provided "as is" without         * 
 *     * express or implied warranty.                                      * 
 *     ********************************************************************* 
 */

#include "global.h"
#include "sim.h"
#include "irsimnet.h"
#include "irsimglobals.h"
#include "irsimnet_macros.h"


typedef struct
{
	float  dynres[2];
	float  rstatic;
} TranResist;


        int     irsim_stack_txtors = FALSE;
static	double  min_cap_ratio = 0;		/* for safe charge sharing */
static	int     nmerged[NTTYPES];		/* number of txtors merged */

void irsim_init_stack(void)
{
	irsim_stack_txtors = FALSE;
}

static void calc_ratio(void)
{
	double  minv;

	minv = 1.0 - irsim_HIGHTHRESH;
	if (irsim_LOWTHRESH < minv)
		minv = irsim_LOWTHRESH;
	min_cap_ratio = (1.0 - minv) / minv;
}


void irsim_pStackedTxtors(void)
{
	int   i, any;

	if (! irsim_stack_txtors)
		return;

	ttyputmsg(x_("stacked transistors:"));
	for(any = FALSE, i = 0; i < NTTYPES; i++)
		if (nmerged[i] != 0)
	{
		ttyputmsg(x_(" %s=%d"), irsim_ttype[i], nmerged[i]);
			any = TRUE;
	}
	ttyputmsg(x_("%s"), (any) ? x_("") : x_(" none"));
}


static int CanMerge(nptr n)
{
	register lptr l;
	register int  type;

	if ((l = n->nterm) == NULL || l->next == NULL)
		return(FALSE);
	type = l->xtor->ttype & ~(GATELIST | ORED);
	l = l->next;
	if (l->next == NULL && type == (l->xtor->ttype & ~(GATELIST | ORED)))
		return(TRUE);
	return(FALSE);
}


#define	NodeQualifies(ND)		\
    (((ND)->ngate == NULL) && CanMerge(ND))



#define	CSproblem(ND, C_SHARE, C_ADJ)			\
    (((ND)->nflags & POWER_RAIL) == 0 &&		\
      ((ND)->ncap - C_ADJ) <= min_cap_ratio * (C_SHARE))



#define	other_tran(T, N)					\
{								\
	register lptr  l;						\
								\
	if ((l = (N)->nterm)->xtor == (T))				\
		l = l->next;						\
	T = l->xtor;						\
}								\


double irsim_StackCap(register tptr t)
{
	register nptr  n;
	double         cap;

	cap = 0.0;
	n = (t->source->nflags & MERGED) ? t->drain : t->source;
	t = (tptr) t->gate;
	do
	{
		n = other_node(t, n);
		cap += n->ncap;
		t = t->scache.t;
	}
	while(t->scache.t != NULL);
	return(cap);
}


#define	GetAnchor(t, n, SHARECAP)				\
{								\
	n = other_node(t, n);					\
	while(NodeQualifies(n))					\
	{								\
		n->nflags &= ~VISITED;					\
		SHARECAP += n->ncap;					\
		other_tran(t, n);					\
		n = other_node(t, n);					\
	}								\
	n->nflags &= ~VISITED;					\
}								\


void irsim_make_stacks(nptr nlist)
{
	register nptr  nd;
	register tptr  t1, t2, stack;
	register nptr  n1, n2;
	register int   tcount;
	double         Cshare, C1, C2;

	if (! irsim_stack_txtors)
		return;

	if (min_cap_ratio == 0)
		calc_ratio();
	for(nd = nlist; nd != NULL; nd->nflags &= ~VISITED, nd = nd->n.next)
	{
		if ((nd->nflags & VISITED) == 0 || NodeQualifies(nd) == FALSE)
			continue;

		n1 = n2 = nd;
		t1 = nd->nterm->xtor;
		t2 = nd->nterm->next->xtor;

		Cshare = nd->ncap;
		C1 = C2 = 0.0;

		GetAnchor(t1, n1, Cshare);
		GetAnchor(t2, n2, Cshare);

		if (t1->ttype & GATELIST)
		{
			C1 = irsim_StackCap(t1) / 2.0;
			Cshare += C1;
		}

		if (t2->ttype & GATELIST)
		{
			C2 = irsim_StackCap(t2) / 2.0;
			Cshare += C2;
		}

		if ((n1 == n2) || (n1->nflags & n2->nflags & POWER_RAIL)
			|| CSproblem(n1, Cshare, C1) || CSproblem(n2, Cshare, C2))
				continue;

		if (n1->nflags & POWER_RAIL)
		{
			SWAP_NODES(n1, n2);
			stack = t1; t1 = t2; t2 = stack;	/* swap t1 & t2 */
		}

		NEW_TRANS(stack);
		stack->r = (Resists *)emalloc(sizeof(TranResist), sim_tool->cluster);
		stack->gate = (t1->ttype & GATELIST) ? t1->gate : (nptr) t1;
		stack->r->rstatic = stack->r->dynhigh = stack->r->dynlow = 0.0;
		stack->ttype = (t1->ttype & ~ORED) | GATELIST;
		stack->state = (t1->ttype & ALWAYSON) ? WEAK : UNKNOWN;
		stack->tflags = 0;
		stack->source = n1;
		stack->drain = n2;

		Cshare /= 2.0;
		n1->ncap += (float)Cshare;
		n2->ncap += (float)Cshare;
		REPLACE(n1->nterm, t1, stack);
		REPLACE(n2->nterm, t2, stack);

		for(tcount = 0;;)
		{
			n1 = other_node(t1, n1);

			stack->r->rstatic += t1->r->rstatic;
			stack->r->dynhigh += t1->r->dynhigh;
			stack->r->dynlow += t1->r->dynlow;

			if (t1->ttype & GATELIST)
			{
				t2 = (tptr) t1->gate;
				for(;;)
				{
					REPLACE(t2->gate->ngate, t2, stack);
					t2->dcache.t = stack;
					if (t2->scache.t == NULL)
						break;
					t2 = t2->scache.t;
				}

				C1 = irsim_StackCap(t1) / 2.0;
				t1->source->ncap -= (float)C1;
				t1->drain->ncap -= (float)C1;

				efree((CHAR *)t1->r);
				FREE_TRANS(t1);
			} else
			{
				tcount++;
				t1->ttype |= STACKED;
				t1->dcache.t = stack;
				REPLACE(t1->gate->ngate, t1, stack);
				t2 = t1;
			}

			if (n1 == n2)
				break;

			other_tran(t1, n1);
			t2->scache.t = (t1->ttype & GATELIST) ? (tptr) t1->gate : t1;

			n1->nterm->next->next = irsim_freeLinks;
			irsim_freeLinks = n1->nterm;
			n1->nterm = NULL;
			n1->t.tran = stack;
			n1->nflags |= MERGED;
			if (n1->curr != &(n1->head))
				irsim_FreeHistList(n1);
		}
		t2->scache.t = NULL;
		nmerged[BASETYPE(t2->ttype)] += tcount;
	}
}


void irsim_DestroyStack(register tptr stack)
{
	register tptr  t;
	register nptr  n;
	register int   tcount;
	double         cap;

	cap = 0.0;
	tcount = 0;
	t = (tptr) stack->gate;
	n = stack->source;
	REPLACE(n->nterm, stack, t);
	n = NULL;
	for(;;)
	{
		if (n == NULL)
			n = stack->source;
		else
		{
			n->nflags &= ~MERGED;
			n->t.cause = NULL;
			cap += n->ncap;
			CONNECT(n->nterm, t);
		}
		REPLACE(t->gate->ngate, stack, t);
		t->ttype &= ~STACKED;
		tcount++;

		n = other_node(t, n);
		if (t->scache.t == NULL)
			break;

		n->nflags &= ~MERGED;
		n->t.cause = NULL;
		CONNECT(n->nterm, t);

		t = t->scache.t;
	}

	REPLACE(n->nterm, stack, t);

	cap /= 2.0;
	stack->source->ncap -= (float)cap;		/* re-adjust capacitance */
	stack->drain->ncap -= (float)cap;

	nmerged[BASETYPE(stack->ttype)] -= tcount;

	efree((CHAR *)stack->r);
	FREE_TRANS(stack);
}

/*
 * Pack NBYTES from NUM into byte stream pointed to by BUFF.
 * Restriction is 1 <= NBYTES <= 4.
 */
void irsim_PackBytes(void *buff, UINTBIG num, int nbytes)
{
	register unsigned long  B_T;
	register UCHAR1         *B_S;

	B_T = num;
	B_S = (UCHAR1 *)buff;

	if (nbytes > 3)
	{
		*B_S++ = (UCHAR1)(B_T & 0x0ff);
		B_T >>= 8;
	}
	if (nbytes > 2)
	{
		*B_S++ = (UCHAR1)(B_T & 0x0ff);
		B_T >>= 8;
	}
	if (nbytes > 1)
	{
		*B_S++ = (UCHAR1)(B_T & 0x0ff);
		B_T >>= 8;
	}
	*B_S = (UCHAR1)(B_T & 0x0ff);
}

/*
 * Extract NBYTES from byte stream BUFF and place result in NUM.
 * Restriction is 1 <= NBYTES <= 4.  (Force Unrolling of loops).
 */
UINTBIG irsim_UnpackBytes(void *buff, int nbytes)
{
	register UCHAR1  *B_S;
	register unsigned long  RES;

	B_S = (UCHAR1 *)buff;
	RES = 0;
	if (nbytes == 1) RES = *B_S;
	else if (nbytes == 2)
	{
		RES = B_S[0];
		RES += (((unsigned int) B_S[1]) << 8) & 0x0ff00;
	}
	else if (nbytes == 3)
	{
		RES = B_S[0];
		RES += (((unsigned int) B_S[1]) << 8) & 0x0ff00;
		RES += (((unsigned int) B_S[2]) << 16) & 0x0ff0000;
	}
	else if (nbytes == 4)
	{
		RES = B_S[0];
		RES += (((unsigned int) B_S[1]) << 8) & 0x0ff00;
		RES += (((unsigned int) B_S[2]) << 16) & 0x0ff0000;
		RES += (((unsigned int) B_S[3]) << 24) & 0xff000000;
	}
	return(RES);
}

/*
 * Pack from TIME into byte stream pointed to by BUFF.
 * Restriction is sizeof(TimeType) == 4 || sizeof(TimeType) == 8
 */
void irsim_PackTime(void *buff, TimeType time)
{
	INTHUGE  B_T;
	register UCHAR1  *B_S;

	B_T = time;
	B_S = (UCHAR1 *)buff;

#ifdef HUGE_TIMETYPE
	*B_S++ = (UCHAR1)(B_T & 0x0ff);
	B_T >>= 8;
	*B_S++ = (UCHAR1)(B_T & 0x0ff);
	B_T >>= 8;
	*B_S++ = (UCHAR1)(B_T & 0x0ff);
	B_T >>= 8;
	*B_S++ = (UCHAR1)(B_T & 0x0ff);
	B_T >>= 8;
#endif
	*B_S++ = (UCHAR1)(B_T & 0x0ff);
	B_T >>= 8;
	*B_S++ = (UCHAR1)(B_T & 0x0ff);
	B_T >>= 8;
	*B_S++ = (UCHAR1)(B_T & 0x0ff);
	B_T >>= 8;
	*B_S = (UCHAR1)(B_T & 0x0ff);
}

/*
 * Extract from byte stream BUFF and place result in TIME.
 * Pack from TIME into byte stream pointed to by BUFF.
 * Restriction is sizeof(TimeType) == 4 || sizeof(TimeType) == 8
 */
TimeType irsim_UnpackTime(void *buff)
{
	register UCHAR1  *B_S;
	register Ulong RES;
#ifdef HUGE_TIMETYPE
	register Ulong RES1;
#endif

	B_S = (UCHAR1 *)buff;
	RES = ((unsigned int) B_S[0]) & 0x0ff;
	RES += (((unsigned int) B_S[1]) << 8) & 0x0ff00;
	RES += (((unsigned int) B_S[2]) << 16) & 0x0ff0000;
	RES += (((unsigned int) B_S[3]) << 24) & 0xff000000;
#ifdef HUGE_TIMETYPE
	RES1 = ((unsigned int) B_S[4]) & 0x0ff;
	RES1 += (((unsigned int) B_S[5]) << 8) & 0x0ff00;
	RES1 += (((unsigned int) B_S[6]) << 16) & 0x0ff0000;
	RES1 += (((unsigned int) B_S[7]) << 24) & 0xff000000;
	return RES + ((INTHUGE)RES1 << 32);
#else
	return(RES);
#endif
}
